from typing import Optional

from ymmsl import SettingValue, Reference, Settings


def has_setting_type(value: SettingValue, typ: str) -> bool:
    """Checks whether the value has the given type.

    Args:
        value: A setting value.
        typ: A setting type. Valid values are 'str', 'int', 'float',
                'bool', '[float]', and '[[float]]'.

    Returns:
        True if the type of value matches typ.

    Raises:
        ValueError: If the type specified is not valid.
    """
    par_type_to_type = {
            'str': str,
            'int': int,
            'float': float,
            'bool': bool
            }

    if typ in par_type_to_type:
        return isinstance(value, par_type_to_type[typ])
    elif typ == '[float]':
        if isinstance(value, list):
            if len(value) == 0 or isinstance(value[0], float):
                # We don't check everything here, the yMMSL loader does
                # a full type check, so we just need to discriminate.
                return True
        return False
    elif typ == '[[float]]':
        if isinstance(value, list):
            if len(value) == 0 or isinstance(value[0], list):
                # We don't check everything here, the yMMSL loader does
                # a full type check, so we just need to discriminate.
                return True
        return False
    raise ValueError('Invalid setting type specified: {}'.format(typ))


class SettingsManager:
    """Manages the current settings for a compute element instance.
    """
    def __init__(self) -> None:
        """Create a SettingsManager.

        Initialises the base and overlay layers to an empty
        Settings object.

        A SettingsManager has two layers of settings, a base
        layer that contains an immutable collection of settings set in
        the simulation's yMMSL description, and an overlay layer that
        holds settings that have been set at run-time.

        Attributes:
            base: The base layer.
            overlay: The overlay layer.
        """
        self.base = Settings()
        self.overlay = Settings()

    def get_setting(self, instance: Reference, setting_name: Reference,
                    typ: Optional[str] = None) -> SettingValue:
        """Returns the value of a setting.

        Args:
            instance: The instance that this value is for.
            setting_name: The name of the setting to get the value of.
            typ: An optional type designation; if specified the type
                    is checked for a match before returning. Valid
                    values are 'str', 'int', 'float', 'bool',
                    '[float]' and '[[float]]'.

        Raises:
            KeyError: If the setting has not been set.
            TypeError: If the setting was set to a value that does
                    not match `typ`.
            ValueError: If an invalid value was specified for `typ`
        """
        for i in range(len(instance), -1, -1):
            if i > 0:
                name = instance[:i] + setting_name
            else:
                name = setting_name

            if name in self.overlay:
                value = self.overlay[name]
                break
            elif name in self.base:
                value = self.base[name]
                break
        else:
            raise KeyError(('Value for setting "{}" was not set.'.format(
                setting_name)))

        if typ is not None:
            if not has_setting_type(value, typ):
                raise TypeError('Value for setting "{}" is of type {},'
                                ' where {} was expected.'.format(
                                    name, type(value), typ))
        return value



import os
import pathlib

from libmuscle.communicator import Message
from libmuscle.instance import Instance
from libmuscle.version import __version__
from libmuscle import runner


# Note that libmuscle.version above is created by the build system; it's okay
# that it's not present.

__all__ = ['__version__', 'Instance', 'Message', 'runner']



from copy import copy
import logging
from pathlib import Path
import sys
from typing import cast, Dict, List, Optional, Tuple, Type, Union

import grpc
from ymmsl import (Conduit, Identifier, Operator, SettingValue, Port,
                   Reference, Settings)

from libmuscle.communicator import Communicator, Message
from libmuscle.settings_manager import SettingsManager
from libmuscle.logging_handler import MuscleManagerHandler
from libmuscle.mcp.message import ClosePort
from libmuscle.mmp_client import MMPClient
from libmuscle.profiler import Profiler
from libmuscle.profiling import ProfileEvent, ProfileEventType
from libmuscle.util import extract_log_file_location


class Instance:
    """Represents a compute element instance in a MUSCLE3 simulation.

    This class provides a low-level send/receive API for the instance
    to use.
    """
    def __init__(self, ports: Optional[Dict[Operator, List[str]]]=None
                 ) -> None:
        """Create an Instance.

        Args:
            ports: A list of port names for each operator of this
                compute element.
        """
        self.__is_shut_down = False

        # Note that these are accessed by Muscle3, but otherwise private.
        self._name, self._index = self.__make_full_name()
        """Name and index of this instance."""

        mmp_location = self.__extract_manager_location()
        self.__manager = MMPClient(mmp_location)
        """Client object for talking to the manager."""

        self.__set_up_logging()

        self._profiler = Profiler(self._instance_name(), self.__manager)
        """Profiler for this instance."""

        self._communicator = Communicator(
                self._name, self._index, ports, self._profiler)
        """Communicator for this instance."""

        self._declared_ports = ports
        """Declared ports for this instance."""

        self._settings_manager = SettingsManager()
        """Settings for this instance."""

        self._first_run = True
        """Keeps track of whether this is the first reuse run."""

        FInitCacheType = Dict[Tuple[str, Optional[int]], Message]
        self._f_init_cache = dict()     # type: FInitCacheType

        self._register()
        self._connect()

    def reuse_instance(self, apply_overlay: bool=True) -> bool:
        """Decide whether to run this instance again.

        In a multiscale simulation, instances get reused all the time.
        For example, in a macro-micro simulation, the micromodel does a
        complete run for every timestep of the macromodel. Rather than
        starting up a new instance of the micromodel, which could be
        expensive, we reuse a single instance many times.

        This may bring other advantages, such as faster convergence
        when starting from the previous final state, and in some
        cases may be necessary if micromodel state needs to be
        preserved from one macro timestep to the next.

        So in MUSCLE, submodels run in a *reuse loop*, which runs them
        over and over again until their work is done and they should be
        shut down. Whether to do another F_INIT, O_I, S, O_F cycle
        is decided by this method.

        This method must be called at the beginning of the reuse loop,
        i.e. before the F_INIT operator, and its return value should
        decide whether to enter that loop again.

        Args:
            apply_overlay: Whether to apply the received settings
                overlay or to save it. If you're going to use
                :meth:`receive_with_settings` on your F_INIT ports,
                set this to False. If you don't know what that means,
                just call `reuse_instance()` without specifying this
                and everything will be fine. If it turns out that you
                did need to specify False, MUSCLE 3 will tell you about
                it in an error message and you can add it still.
        """
        do_reuse = self.__receive_settings()

        # TODO: _f_init_cache should be empty here, or the user didn't
        # receive something that was sent on the last go-around.
        # At least emit a warning.
        self.__pre_receive_f_init(apply_overlay)

        ports = self._communicator.list_ports()
        f_init_not_connected = all(
                [not self.is_connected(port)
                 for port in ports.get(Operator.F_INIT, [])])
        no_settings_in = not self._communicator.settings_in_connected()

        if f_init_not_connected and no_settings_in:
            do_reuse = self._first_run
            self._first_run = False
        else:
            for message in self._f_init_cache.values():
                if isinstance(message.data, ClosePort):
                    do_reuse = False

        if not do_reuse:
            self.__close_ports()
            self._communicator.shutdown()
            self._deregister()
        return do_reuse

    def error_shutdown(self, message: str) -> None:
        """Logs an error and shuts down the Instance.

        If you detect that something is wrong (invalid input, invalid
        settings, simulation diverged, or anything else really), you
        should call this method before calling exit() or raising
        an exception that you don't expect to catch.

        If you do so, the Instance will tell the rest of the simulation
        that it encountered an error and will shut down. That makes it
        easier to debug the situation (the message will be logged), and
        it reduces the chance that other parts of the simulation will
        sit around waiting forever for a message that this instance was
        supposed to send.

        Args:
            message: An error message describing the problem.
        """
        self.__shutdown(message)

    def get_setting(self, name: str, typ: Optional[str] = None
                    ) -> SettingValue:
        """Returns the value of a model setting.

        Args:
            name: The name of the setting, without any instance
                    prefix.
            typ: The expected type of the value. If the value does
                    not match this type, a TypeError will be raised.
                    If not specified, any of the supported types
                    will be accepted, and you'll have to figure out
                    what you got yourself.

        Raises:
            KeyError: If no value was set for this setting.
            TypeError: If the type of the setting's value was not
                    as expected.
        """
        return self._settings_manager.get_setting(
                self._instance_name(), Reference(name), typ)

    def list_ports(self) -> Dict[Operator, List[str]]:
        """Returns a description of the ports that this CE has.

        Note that the result has almost the same format as the port
        declarations you pass when making an Instance. The only
        difference is that the port names never have `[]` at the end,
        even if the port is a vector port.

        Returns:
            A dictionary, indexed by Operator, containing lists of
            port names. Operators with no associated ports are not
            included.
        """
        return self._communicator.list_ports()

    def is_connected(self, port: str) -> bool:
        """Returns whether the given port is connected.

        Args:
            port: The name of the port to inspect.

        Returns:
            True if there is a conduit attached to this port, False
            if not.
        """
        return self._communicator.get_port(port).is_connected()

    def is_vector_port(self, port: str) -> bool:
        """Returns whether a port is a vector or scalar port

        If a port has been declared to be a vector port (i.e. the
        name passed when creating this Instance had '[]' at the
        end), then you can pass a 'slot' argument when sending or
        receiving. It's like the port is a vector of slots on which
        you can send or receive messages.

        This function returns True if the given port is a vector
        port, and False if it is a scalar port.

        Args:
            port: The port to check this property of.
        """
        return self._communicator.get_port(port).is_vector()

    def is_resizable(self, port: str) -> bool:
        """Returns whether the given port is resizable.

        Scalar ports are never resizable. Whether a vector port is
        resizable depends on what it is connected to.

        Args:
            port: Name of the port to inspect.

        Returns:
            True if the port can be resized.
        """
        return self._communicator.get_port(port).is_resizable()

    def get_port_length(self, port: str) -> int:
        """Returns the current length of the port.

        Args:
            port: The name of the port to measure.

        Raises: RuntimeError if this is a scalar port.
        """
        return self._communicator.get_port(port).get_length()

    def set_port_length(self, port: str, length: int) -> None:
        """Resizes the port to the given length.

        You should check whether the port is resizable using
        `is_resizable()` first; whether it is depends on how this
        compute element is wired up, so you should check.

        Args:
            port: Name of the port to resize.
            length: The new length.

        Raises:
            RuntimeError: If the port is not resizable.
        """
        self._communicator.get_port(port).set_length(length)

    def send(self, port_name: str, message: Message,
             slot: Optional[int]=None) -> None:
        """Send a message to the outside world.

        Sending is non-blocking, a copy of the message will be made
        and stored until the receiver is ready to receive it.

        Args:
            port_name: The port on which this message is to be sent.
            message: The message to be sent.
            slot: The slot to send the message on, if any.
        """
        self.__check_port(port_name)
        if message.settings is None:
            message = copy(message)
            message.settings = self._settings_manager.overlay

        self._communicator.send_message(port_name, message, slot)

    def receive(self, port_name: str, slot: Optional[int]=None,
                default: Optional[Message]=None
                ) -> Message:
        """Receive a message from the outside world.

        Receiving is a blocking operation. This function will contact
        the sender, wait for a message to be available, and receive and
        return it.

        If the port you are receiving on is not connected, the default
        value you specified will be returned exactly as you passed it.
        If you didn't specify a default value (e.g. because there is no
        reasonable default, you really need the outside input) and the
        port is not connected, you'll get a RuntimeError.

        Args:
            port_name: The endpoint on which a message is to be
                    received.
            slot: The slot to receive the message on, if any.
            default: A default value to return if this port is not
                    connected.

        Returns:
            The received message. The settings attribute of the
            received message will be None.

        Raises:
            RuntimeError: If the given port is not connected and no
                    default value was given.
        """
        return self.__receive_message(port_name, slot, default, False)

    def receive_with_settings(
            self, port_name: str, slot: Optional[int]=None,
            default: Optional[Message]=None
            ) -> Message:
        """Receive a message with attached settings overlay.

        This function should not be used in submodels. It is intended
        for use by special compute elements that are ensemble-aware and
        have to pass on overlay settings explicitly.

        Receiving is a blocking operation. This function will contact
        the sender, wait for a message to be available, and receive and
        return it.

        If the port you are receiving on is not connected, the default
        value you specified will be returned exactly as you passed it.
        If you didn't specify a default value (e.g. because there is no
        reasonable default, and you really need the outside input) and
        the port is not connected, then you'll get a RuntimeError.

        Args:
            port_name: The endpoint on which a message is to be
                    received.
            slot: The slot to receive the message on, if any.
            default: A default value to return if this port is not
                    connected.

        Returns:
            The received message. The settings attribute will
            contain the received Settings, and will not be None.

        Raises:
            RuntimeError: If the given port is not connected and no
                    default value was given.
        """
        return self.__receive_message(port_name, slot, default, True)

    def _register(self) -> None:
        """Register this instance with the manager.
        """
        register_event = self._profiler.start(ProfileEventType.REGISTER)
        locations = self._communicator.get_locations()
        port_list = self.__list_declared_ports()
        self.__manager.register_instance(self._instance_name(), locations,
                                         port_list)
        try:
            register_event.stop()
        except grpc._channel._Rendezvous:
            # This may happen if we're the last submodel to quit, and the
            # manager is already gone. Nothing we can do in that case, and this
            # final Register event will be lost, which is not a big issue.
            pass

    def _connect(self) -> None:
        """Connect this instance to the given peers / conduits.
        """
        connect_event = self._profiler.start(ProfileEventType.CONNECT)
        conduits, peer_dims, peer_locations = self.__manager.request_peers(
                self._instance_name())
        self._communicator.connect(conduits, peer_dims, peer_locations)
        self._settings_manager.base = self.__manager.get_settings()
        connect_event.stop()

    def _deregister(self) -> None:
        """Deregister this instance from the manager.
        """
        deregister_event = self._profiler.start(ProfileEventType.DEREGISTER)
        self.__manager.deregister_instance(self._instance_name())
        deregister_event.stop()
        # this is the last thing we'll profile, so flush messages
        self._profiler.shutdown()

    @staticmethod
    def __extract_manager_location() -> str:
        """Gets the manager network location from the command line.

        We use a --muscle-manager=<host:port> argument to tell the
        MUSCLE library how to connect to the manager. This function
        will extract this argument from the command line arguments,
        if it is present.

        Returns:
            A connection string, or None.
        """
        # Neither getopt, optparse, or argparse will let me pick out
        # just one option from the command line and ignore the rest.
        # So we do it by hand.
        prefix = '--muscle-manager='
        for arg in sys.argv[1:]:
            if arg.startswith(prefix):
                return arg[len(prefix):]

        return 'localhost:9000'

    def __set_up_logging(self) -> None:
        """Adds logging handlers for one or more instances.
        """
        id_str = str(self._instance_name())

        logfile = extract_log_file_location('muscle3.{}.log'.format(id_str))
        local_handler = logging.FileHandler(str(logfile), mode='w')
        formatter = logging.Formatter('%(asctime)-15s: %(name)s'
                                      ' %(levelname)s: %(message)s')
        local_handler.setFormatter(formatter)
        logging.getLogger().addHandler(local_handler)

        if self.__manager is not None:
            mmp_handler = MuscleManagerHandler(id_str, logging.WARNING,
                                               self.__manager)
            logging.getLogger().addHandler(mmp_handler)

    def __receive_message(
            self, port_name: str, slot: Optional[int],
            default: Optional[Message], with_settings: bool
            ) -> Message:
        """Receives a message on the given port.

        This implements receive and receive_with_settings, see the
        description of those.
        """
        self.__check_port(port_name)

        port = self._communicator.get_port(port_name)
        if port.operator == Operator.F_INIT:
            if (port_name, slot) in self._f_init_cache:
                msg = self._f_init_cache[(port_name, slot)]
                del(self._f_init_cache[(port_name, slot)])
                if with_settings and msg.settings is None:
                    err_msg = ('If you use receive_with_settings()'
                               ' on an F_INIT port, then you have to'
                               ' pass False to reuse_instance(),'
                               ' otherwise the settings will already'
                               ' have been applied by MUSCLE.')
                    self.__shutdown(err_msg)
                    raise RuntimeError(err_msg)
            else:
                if port.is_connected():
                    err_msg = (('Tried to receive twice on the same'
                                ' port "{}", that\'s not possible.'
                                ' Did you forget to call'
                                ' reuse_instance() in your reuse loop?'
                                ).format(port_name))
                    self.__shutdown(err_msg)
                    raise RuntimeError(err_msg)
                else:
                    if default is not None:
                        return default
                    err_msg = (('Tried to receive on port "{}",'
                                ' which is not connected, and no'
                                ' default value was given. Please'
                                ' connect this port!').format(port_name))
                    self.__shutdown(err_msg)
                    raise RuntimeError(err_msg)

        else:
            msg = self._communicator.receive_message(
                    port_name, slot, default)
            if port.is_connected and not port.is_open(slot):
                err_msg = (('Port {} was closed while trying to'
                            ' receive on it, did the peer crash?'
                            ).format(port_name))
                self.__shutdown(err_msg)
                raise RuntimeError(err_msg)
            if port.is_connected and not with_settings:
                self.__check_compatibility(port_name, msg.settings)
            if not with_settings:
                msg.settings = None
        return msg

    def __make_full_name(self
                         ) -> Tuple[Reference, List[int]]:
        """Returns instance name and index.

        This takes the argument to the --muscle-instance= command-line
        option and splits it into a compute element name and an index.
        """
        def split_reference(ref: Reference) -> Tuple[Reference, List[int]]:
            index = list()     # type: List[int]
            i = 0
            while i < len(ref) and isinstance(ref[i], Identifier):
                i += 1
            name = ref[:i]

            while i < len(ref) and isinstance(ref[i], int):
                index.append(cast(int, ref[i]))
                i += 1

            return name, index

        # Neither getopt, optparse, or argparse will let me pick out
        # just one option from the command line and ignore the rest.
        # So we do it by hand.
        prefix_tag = '--muscle-instance='
        for arg in sys.argv[1:]:
            if arg.startswith(prefix_tag):
                prefix_str = arg[len(prefix_tag):]
                prefix_ref = Reference(prefix_str)
                name, index = split_reference(prefix_ref)
                break
        else:
            raise RuntimeError(('A --muscle-instance command line argument is'
                                ' required to identify this instance. Please'
                                ' add one.'))
        return name, index

    def __list_declared_ports(self) -> List[Port]:
        """Returns a list of declared ports.

        This returns a list of ymmsl.Port objects, which have only the
        name and the operator, not libmuscle.Port, which has more.
        """
        result = list()
        if self._declared_ports is not None:
            for operator, port_names in self._declared_ports.items():
                for name in port_names:
                    if name.endswith('[]'):
                        name = name[:-2]
                    result.append(Port(Identifier(name), operator))
        return result

    def _instance_name(self) -> Reference:
        """Returns the full instance name.
        """
        return self._name + self._index

    def __check_port(self, port_name: str) -> None:
        if not self._communicator.port_exists(port_name):
            err_msg = (('Port "{}" does not exist on "{}". Please check'
                        ' the name and the list of ports you gave for'
                        ' this compute element.').format(port_name,
                                                         self._name))
            self.__shutdown(err_msg)
            raise RuntimeError(err_msg)

    def __receive_settings(self) -> bool:
        """Receives settings on muscle_settings_in.

        Returns:
            False iff the port is connnected and ClosePort was received.
        """
        default_message = Message(0.0, None, Settings(), Settings())
        message = self._communicator.receive_message(
                'muscle_settings_in', None, default_message)
        if isinstance(message.data, ClosePort):
            return False
        if not isinstance(message.data, Settings):
            err_msg = ('"{}" received a message on'
                       ' muscle_settings_in that is not a'
                       ' Settings. It seems that your'
                       ' simulation is miswired or the sending'
                       ' instance is broken.'.format(self._instance_name()))
            self.__shutdown(err_msg)
            raise RuntimeError(err_msg)

        settings = cast(Settings, message.settings)
        for key, value in message.data.items():
            settings[key] = value
        self._settings_manager.overlay = settings
        return True

    def __pre_receive_f_init(self, apply_overlay: bool) -> None:
        """Receives on all ports connected to F_INIT.

        This receives all incoming messages on F_INIT and stores them
        in self._f_init_cache.
        """
        def pre_receive(port_name: str, slot: Optional[int]) -> None:
            msg = self._communicator.receive_message(port_name, slot)
            self._f_init_cache[(port_name, slot)] = msg
            if apply_overlay:
                self.__apply_overlay(msg)
                self.__check_compatibility(port_name, msg.settings)
                msg.settings = None

        self._f_init_cache = dict()
        ports = self._communicator.list_ports()
        for port_name in ports.get(Operator.F_INIT, []):
            port = self._communicator.get_port(port_name)
            if not port.is_connected():
                continue
            if not port.is_vector():
                pre_receive(port_name, None)
            else:
                pre_receive(port_name, 0)
                # The above receives the length, if needed, so now we can
                # get the rest.
                for slot in range(1, port.get_length()):
                    pre_receive(port_name, slot)

    def __apply_overlay(self, message: Message) -> None:
        """Sets local overlay if we don't already have one.

        Args:
            message: The message to apply the overlay from.
        """
        if len(self._settings_manager.overlay) == 0:
            if message.settings is not None:
                self._settings_manager.overlay = message.settings

    def __check_compatibility(self, port_name: str,
                              overlay: Optional[Settings]) -> None:
        """Checks whether a received overlay matches the current one.

        Args:
            port_name: Name of the port on which the overlay was
                    received.
            overlay: The received overlay.
        """
        if overlay is None:
            return
        if self._settings_manager.overlay != overlay:
            err_msg = (('Unexpectedly received data from a'
                        ' parallel universe on port "{}". My'
                        ' settings are "{}" and I received'
                        ' from a universe with "{}".').format(
                            port_name, self._settings_manager.overlay,
                            overlay))
            self.__shutdown(err_msg)
            raise RuntimeError(err_msg)

    def __close_outgoing_ports(self) -> None:
        """Closes outgoing ports.

        This sends a close port message on all slots of all outgoing
        ports.
        """
        for operator, ports in self._communicator.list_ports().items():
            if operator.allows_sending():
                for port_name in ports:
                    port = self._communicator.get_port(port_name)
                    if port.is_vector():
                        for slot in range(port.get_length()):
                            self._communicator.close_port(port_name, slot)
                    else:
                        self._communicator.close_port(port_name)

    def __drain_incoming_port(self, port_name: str) -> None:
        """Receives messages until a ClosePort is received.

        Receives at least once.

        Args:
            port_name: Port to drain.
        """
        port = self._communicator.get_port(port_name)
        while port.is_open():
            # TODO: log warning if not a ClosePort
            self._communicator.receive_message(port_name)

    def __drain_incoming_vector_port(self, port_name: str) -> None:
        """Receives messages until a ClosePort is received.

        Works with (resizable) vector ports.

        Args:
            port_name: Port to drain.
        """
        port = self._communicator.get_port(port_name)
        while not all([not port.is_open(slot)
                       for slot in range(port.get_length())]):
            for slot in range(port.get_length()):
                if port.is_open(slot):
                    self._communicator.receive_message(port_name, slot)

    def __close_incoming_ports(self) -> None:
        """Closes incoming ports.

        This receives on all incoming ports until a ClosePort is
        received on them, signaling that there will be no more
        messages, and allowing the sending instance to shut down
        cleanly.
        """
        for operator, port_names in self._communicator.list_ports().items():
            if operator.allows_receiving():
                for port_name in port_names:
                    port = self._communicator.get_port(port_name)
                    if not port.is_connected():
                        continue
                    if not port.is_vector():
                        self.__drain_incoming_port(port_name)
                    else:
                        self.__drain_incoming_vector_port(port_name)

    def __close_ports(self) -> None:
        """Closes all ports.

        This sends a close port message on all slots of all outgoing
        ports, then receives one on all incoming ports.
        """
        self.__close_outgoing_ports()
        self.__close_incoming_ports()

    def __shutdown(self, message: str) -> None:
        """Shuts down simulation.

        This logs the given error message, communicates to the peers
        that we're shutting down, and deregisters from the manager.
        """
        if not self.__is_shut_down:
            logging.critical(message)
            self.__close_ports()
            self._communicator.shutdown()
            self._deregister()
            self.__is_shut_down = True



import socketserver as ss
import threading
from typing import cast, List, Optional, Tuple, Type

import msgpack
import netifaces
from ymmsl import Reference

from libmuscle.mcp.server import Server
from libmuscle.mcp.tcp_util import (recv_all, recv_int64, send_int64,
                                    SocketClosed)
from libmuscle.post_office import PostOffice


class TcpServerImpl(ss.ThreadingMixIn, ss.TCPServer):
    def __init__(self, host_port_tuple: Tuple[str, int],
                 streamhandler: Type, tcp_server: 'TcpServer'
                 ) -> None:
        super().__init__(host_port_tuple, streamhandler)
        self.tcp_server = tcp_server


class TcpHandler(ss.BaseRequestHandler):
    """Handler for MCP-over-TCP connections.
    """
    def handle(self) -> None:
        """Handles requests on a socket
        """
        receiver_id = self.receive_request()

        while receiver_id is not None:
            server = cast(TcpServerImpl, self.server).tcp_server
            message = server.post_office.get_message(receiver_id)

            send_int64(self.request, len(message))
            self.request.sendall(message)
            receiver_id = self.receive_request()

    def receive_request(self) -> Optional[Reference]:
        """Receives a request (receiver id).

        Returns:
            The received receiver id.
        """
        try:
            length = recv_int64(self.request)
            reqbuf = recv_all(self.request, length)
            return Reference(reqbuf.decode('utf-8'))
        except SocketClosed:
            return None


class TcpServer(Server):
    """A server that accepts MCP connections over TCP.
    """
    def __init__(self, instance_id: Reference, post_office: PostOffice
                 ) -> None:
        """Create a TCPServer.

        Args:
            instance_id: Id of the instance we're a server for.
            post_office: A PostOffice to obtain data from.
        """
        super().__init__(instance_id, post_office)

        self._server = TcpServerImpl(('', 0), TcpHandler, self)
        self._server_thread = threading.Thread(
                target=self._server.serve_forever)
        self._server_thread.start()

    def get_location(self) -> str:
        """Returns the location this server listens on.

        Returns:
            A string containing the location.
        """
        host, port = self._server.server_address

        locs = list()   # type: List[str]
        for address in self._get_if_addresses():
            locs.append('{}:{}'.format(address, port))
        return 'tcp:{}'.format(','.join(locs))

    def close(self) -> None:
        """Closes this server.

        Stops the server listening, waits for existing clients to
        disconnect, then frees any other resources.
        """
        self._server.shutdown()
        self._server_thread.join()

    @property
    def post_office(self) -> PostOffice:
        """Export this so the server thread can use it.
        """
        return self._post_office

    def _get_if_addresses(self) -> List[str]:
        all_addresses = list()  # type: List[str]
        ifs = netifaces.interfaces()
        for interface in ifs:
            addrs = netifaces.ifaddresses(interface)
            for props in addrs.get(netifaces.AF_INET, []):
                all_addresses.append(props['addr'])
            for props in addrs.get(netifaces.AF_INET6, []):
                # filter out link-local addresses with a scope id
                if '%' not in props['addr']:
                    all_addresses.append('[' + props['addr'] + ']')
        return all_addresses



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: muscle_manager_protocol.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='muscle_manager_protocol.proto',
  package='muscle_manager_protocol',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x1dmuscle_manager_protocol.proto\x12\x17muscle_manager_protocol\x1a\x1fgoogle/protobuf/timestamp.proto\"\x90\x01\n\nLogMessage\x12\x13\n\x0binstance_id\x18\x01 \x01(\t\x12-\n\ttimestamp\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x30\n\x05level\x18\x04 \x01(\x0e\x32!.muscle_manager_protocol.LogLevel\x12\x0c\n\x04text\x18\x05 \x01(\t\"\x0b\n\tLogResult\"I\n\x04Port\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x33\n\x08operator\x18\x02 \x01(\x0e\x32!.muscle_manager_protocol.Operator\"\xa7\x02\n\x0cProfileEvent\x12\x13\n\x0binstance_id\x18\x01 \x01(\t\x12.\n\nstart_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12-\n\tstop_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12=\n\nevent_type\x18\x04 \x01(\x0e\x32).muscle_manager_protocol.ProfileEventType\x12+\n\x04port\x18\x05 \x01(\x0b\x32\x1d.muscle_manager_protocol.Port\x12\x13\n\x0bport_length\x18\x06 \x01(\x05\x12\x0c\n\x04slot\x18\x07 \x01(\x05\x12\x14\n\x0cmessage_size\x18\x08 \x01(\x03\"@\n\x07Profile\x12\x35\n\x06\x65vents\x18\x01 \x03(\x0b\x32%.muscle_manager_protocol.ProfileEvent\"\x0f\n\rProfileResult\"u\n\x13RegistrationRequest\x12\x15\n\rinstance_name\x18\x01 \x01(\t\x12\x19\n\x11network_locations\x18\x02 \x03(\t\x12,\n\x05ports\x18\x03 \x03(\x0b\x32\x1d.muscle_manager_protocol.Port\"b\n\x12RegistrationResult\x12\x35\n\x06status\x18\x01 \x01(\x0e\x32%.muscle_manager_protocol.ResultStatus\x12\x15\n\rerror_message\x18\x02 \x01(\t\"+\n\x07\x43onduit\x12\x0e\n\x06sender\x18\x01 \x01(\t\x12\x10\n\x08receiver\x18\x02 \x01(\t\"$\n\x0bPeerRequest\x12\x15\n\rinstance_name\x18\x01 \x01(\t\"\x9a\x03\n\nPeerResult\x12\x35\n\x06status\x18\x01 \x01(\x0e\x32%.muscle_manager_protocol.ResultStatus\x12\x15\n\rerror_message\x18\x02 \x01(\t\x12\x32\n\x08\x63onduits\x18\x03 \x03(\x0b\x32 .muscle_manager_protocol.Conduit\x12K\n\x0fpeer_dimensions\x18\x04 \x03(\x0b\x32\x32.muscle_manager_protocol.PeerResult.PeerDimensions\x12I\n\x0epeer_locations\x18\x05 \x03(\x0b\x32\x31.muscle_manager_protocol.PeerResult.PeerLocations\x1a\x37\n\x0ePeerDimensions\x12\x11\n\tpeer_name\x18\x01 \x01(\t\x12\x12\n\ndimensions\x18\x02 \x03(\x05\x1a\x39\n\rPeerLocations\x12\x15\n\rinstance_name\x18\x01 \x01(\t\x12\x11\n\tlocations\x18\x02 \x03(\t\"\x1e\n\x0cListOfDouble\x12\x0e\n\x06values\x18\x01 \x03(\x01\"K\n\x12ListOfListOfDouble\x12\x35\n\x06values\x18\x01 \x03(\x0b\x32%.muscle_manager_protocol.ListOfDouble\"\xb5\x02\n\x07Setting\x12\x0c\n\x04name\x18\x01 \x01(\t\x12=\n\nvalue_type\x18\x02 \x01(\x0e\x32).muscle_manager_protocol.SettingValueType\x12\x14\n\x0cvalue_string\x18\x03 \x01(\t\x12\x11\n\tvalue_int\x18\x04 \x01(\x03\x12\x13\n\x0bvalue_float\x18\x05 \x01(\x01\x12\x12\n\nvalue_bool\x18\x06 \x01(\x08\x12?\n\x10value_list_float\x18\x07 \x01(\x0b\x32%.muscle_manager_protocol.ListOfDouble\x12J\n\x15value_list_list_float\x18\x08 \x01(\x0b\x32+.muscle_manager_protocol.ListOfListOfDouble\"\x11\n\x0fSettingsRequest\"J\n\x0eSettingsResult\x12\x38\n\x0esetting_values\x18\x01 \x03(\x0b\x32 .muscle_manager_protocol.Setting\".\n\x15\x44\x65registrationRequest\x12\x15\n\rinstance_name\x18\x01 \x01(\t\"d\n\x14\x44\x65registrationResult\x12\x35\n\x06status\x18\x01 \x01(\x0e\x32%.muscle_manager_protocol.ResultStatus\x12\x15\n\rerror_message\x18\x02 \x01(\t*]\n\x0cResultStatus\x12\x19\n\x15RESULT_STATUS_SUCCESS\x10\x00\x12\x17\n\x13RESULT_STATUS_ERROR\x10\x01\x12\x19\n\x15RESULT_STATUS_PENDING\x10\x02*\x88\x01\n\x08Operator\x12\x11\n\rOPERATOR_NONE\x10\x00\x12\x13\n\x0fOPERATOR_F_INIT\x10\x01\x12\x10\n\x0cOPERATOR_O_I\x10\x02\x12\x0e\n\nOPERATOR_S\x10\x03\x12\x0e\n\nOPERATOR_B\x10\x04\x12\x10\n\x0cOPERATOR_O_F\x10\x05\x12\x10\n\x0cOPERATOR_MAP\x10\x06*w\n\x08LogLevel\x12\x13\n\x0fLOG_LEVEL_DEBUG\x10\x00\x12\x12\n\x0eLOG_LEVEL_INFO\x10\x01\x12\x15\n\x11LOG_LEVEL_WARNING\x10\x03\x12\x13\n\x0fLOG_LEVEL_ERROR\x10\x04\x12\x16\n\x12LOG_LEVEL_CRITICAL\x10\x05*\xb3\x01\n\x10ProfileEventType\x12\x1f\n\x1bPROFILE_EVENT_TYPE_REGISTER\x10\x00\x12\x1e\n\x1aPROFILE_EVENT_TYPE_CONNECT\x10\x04\x12!\n\x1dPROFILE_EVENT_TYPE_DEREGISTER\x10\x01\x12\x1b\n\x17PROFILE_EVENT_TYPE_SEND\x10\x02\x12\x1e\n\x1aPROFILE_EVENT_TYPE_RECEIVE\x10\x03*\xd3\x01\n\x10SettingValueType\x12\x1d\n\x19SETTING_VALUE_TYPE_STRING\x10\x00\x12\x1a\n\x16SETTING_VALUE_TYPE_INT\x10\x01\x12\x1c\n\x18SETTING_VALUE_TYPE_FLOAT\x10\x02\x12\x1b\n\x17SETTING_VALUE_TYPE_BOOL\x10\x03\x12!\n\x1dSETTING_VALUE_TYPE_LIST_FLOAT\x10\x04\x12&\n\"SETTING_VALUE_TYPE_LIST_LIST_FLOAT\x10\x05\x32\xf2\x04\n\rMuscleManager\x12[\n\x10SubmitLogMessage\x12#.muscle_manager_protocol.LogMessage\x1a\".muscle_manager_protocol.LogResult\x12_\n\x13SubmitProfileEvents\x12 .muscle_manager_protocol.Profile\x1a&.muscle_manager_protocol.ProfileResult\x12\x64\n\x0fRequestSettings\x12(.muscle_manager_protocol.SettingsRequest\x1a\'.muscle_manager_protocol.SettingsResult\x12m\n\x10RegisterInstance\x12,.muscle_manager_protocol.RegistrationRequest\x1a+.muscle_manager_protocol.RegistrationResult\x12Y\n\x0cRequestPeers\x12$.muscle_manager_protocol.PeerRequest\x1a#.muscle_manager_protocol.PeerResult\x12s\n\x12\x44\x65registerInstance\x12..muscle_manager_protocol.DeregistrationRequest\x1a-.muscle_manager_protocol.DeregistrationResultb\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,])

_RESULTSTATUS = _descriptor.EnumDescriptor(
  name='ResultStatus',
  full_name='muscle_manager_protocol.ResultStatus',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='RESULT_STATUS_SUCCESS', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RESULT_STATUS_ERROR', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RESULT_STATUS_PENDING', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2088,
  serialized_end=2181,
)
_sym_db.RegisterEnumDescriptor(_RESULTSTATUS)

ResultStatus = enum_type_wrapper.EnumTypeWrapper(_RESULTSTATUS)
_OPERATOR = _descriptor.EnumDescriptor(
  name='Operator',
  full_name='muscle_manager_protocol.Operator',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='OPERATOR_NONE', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OPERATOR_F_INIT', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OPERATOR_O_I', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OPERATOR_S', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OPERATOR_B', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OPERATOR_O_F', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OPERATOR_MAP', index=6, number=6,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2184,
  serialized_end=2320,
)
_sym_db.RegisterEnumDescriptor(_OPERATOR)

Operator = enum_type_wrapper.EnumTypeWrapper(_OPERATOR)
_LOGLEVEL = _descriptor.EnumDescriptor(
  name='LogLevel',
  full_name='muscle_manager_protocol.LogLevel',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='LOG_LEVEL_DEBUG', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOG_LEVEL_INFO', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOG_LEVEL_WARNING', index=2, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOG_LEVEL_ERROR', index=3, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOG_LEVEL_CRITICAL', index=4, number=5,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2322,
  serialized_end=2441,
)
_sym_db.RegisterEnumDescriptor(_LOGLEVEL)

LogLevel = enum_type_wrapper.EnumTypeWrapper(_LOGLEVEL)
_PROFILEEVENTTYPE = _descriptor.EnumDescriptor(
  name='ProfileEventType',
  full_name='muscle_manager_protocol.ProfileEventType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='PROFILE_EVENT_TYPE_REGISTER', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PROFILE_EVENT_TYPE_CONNECT', index=1, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PROFILE_EVENT_TYPE_DEREGISTER', index=2, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PROFILE_EVENT_TYPE_SEND', index=3, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PROFILE_EVENT_TYPE_RECEIVE', index=4, number=3,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2444,
  serialized_end=2623,
)
_sym_db.RegisterEnumDescriptor(_PROFILEEVENTTYPE)

ProfileEventType = enum_type_wrapper.EnumTypeWrapper(_PROFILEEVENTTYPE)
_SETTINGVALUETYPE = _descriptor.EnumDescriptor(
  name='SettingValueType',
  full_name='muscle_manager_protocol.SettingValueType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SETTING_VALUE_TYPE_STRING', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SETTING_VALUE_TYPE_INT', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SETTING_VALUE_TYPE_FLOAT', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SETTING_VALUE_TYPE_BOOL', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SETTING_VALUE_TYPE_LIST_FLOAT', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SETTING_VALUE_TYPE_LIST_LIST_FLOAT', index=5, number=5,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2626,
  serialized_end=2837,
)
_sym_db.RegisterEnumDescriptor(_SETTINGVALUETYPE)

SettingValueType = enum_type_wrapper.EnumTypeWrapper(_SETTINGVALUETYPE)
RESULT_STATUS_SUCCESS = 0
RESULT_STATUS_ERROR = 1
RESULT_STATUS_PENDING = 2
OPERATOR_NONE = 0
OPERATOR_F_INIT = 1
OPERATOR_O_I = 2
OPERATOR_S = 3
OPERATOR_B = 4
OPERATOR_O_F = 5
OPERATOR_MAP = 6
LOG_LEVEL_DEBUG = 0
LOG_LEVEL_INFO = 1
LOG_LEVEL_WARNING = 3
LOG_LEVEL_ERROR = 4
LOG_LEVEL_CRITICAL = 5
PROFILE_EVENT_TYPE_REGISTER = 0
PROFILE_EVENT_TYPE_CONNECT = 4
PROFILE_EVENT_TYPE_DEREGISTER = 1
PROFILE_EVENT_TYPE_SEND = 2
PROFILE_EVENT_TYPE_RECEIVE = 3
SETTING_VALUE_TYPE_STRING = 0
SETTING_VALUE_TYPE_INT = 1
SETTING_VALUE_TYPE_FLOAT = 2
SETTING_VALUE_TYPE_BOOL = 3
SETTING_VALUE_TYPE_LIST_FLOAT = 4
SETTING_VALUE_TYPE_LIST_LIST_FLOAT = 5



_LOGMESSAGE = _descriptor.Descriptor(
  name='LogMessage',
  full_name='muscle_manager_protocol.LogMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='instance_id', full_name='muscle_manager_protocol.LogMessage.instance_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='muscle_manager_protocol.LogMessage.timestamp', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='level', full_name='muscle_manager_protocol.LogMessage.level', index=2,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='text', full_name='muscle_manager_protocol.LogMessage.text', index=3,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=92,
  serialized_end=236,
)


_LOGRESULT = _descriptor.Descriptor(
  name='LogResult',
  full_name='muscle_manager_protocol.LogResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=238,
  serialized_end=249,
)


_PORT = _descriptor.Descriptor(
  name='Port',
  full_name='muscle_manager_protocol.Port',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='muscle_manager_protocol.Port.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='operator', full_name='muscle_manager_protocol.Port.operator', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=251,
  serialized_end=324,
)


_PROFILEEVENT = _descriptor.Descriptor(
  name='ProfileEvent',
  full_name='muscle_manager_protocol.ProfileEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='instance_id', full_name='muscle_manager_protocol.ProfileEvent.instance_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='muscle_manager_protocol.ProfileEvent.start_time', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stop_time', full_name='muscle_manager_protocol.ProfileEvent.stop_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='event_type', full_name='muscle_manager_protocol.ProfileEvent.event_type', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='port', full_name='muscle_manager_protocol.ProfileEvent.port', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='port_length', full_name='muscle_manager_protocol.ProfileEvent.port_length', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='slot', full_name='muscle_manager_protocol.ProfileEvent.slot', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message_size', full_name='muscle_manager_protocol.ProfileEvent.message_size', index=7,
      number=8, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=327,
  serialized_end=622,
)


_PROFILE = _descriptor.Descriptor(
  name='Profile',
  full_name='muscle_manager_protocol.Profile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='events', full_name='muscle_manager_protocol.Profile.events', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=624,
  serialized_end=688,
)


_PROFILERESULT = _descriptor.Descriptor(
  name='ProfileResult',
  full_name='muscle_manager_protocol.ProfileResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=690,
  serialized_end=705,
)


_REGISTRATIONREQUEST = _descriptor.Descriptor(
  name='RegistrationRequest',
  full_name='muscle_manager_protocol.RegistrationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='instance_name', full_name='muscle_manager_protocol.RegistrationRequest.instance_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='network_locations', full_name='muscle_manager_protocol.RegistrationRequest.network_locations', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ports', full_name='muscle_manager_protocol.RegistrationRequest.ports', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=707,
  serialized_end=824,
)


_REGISTRATIONRESULT = _descriptor.Descriptor(
  name='RegistrationResult',
  full_name='muscle_manager_protocol.RegistrationResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='muscle_manager_protocol.RegistrationResult.status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error_message', full_name='muscle_manager_protocol.RegistrationResult.error_message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=826,
  serialized_end=924,
)


_CONDUIT = _descriptor.Descriptor(
  name='Conduit',
  full_name='muscle_manager_protocol.Conduit',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='sender', full_name='muscle_manager_protocol.Conduit.sender', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='receiver', full_name='muscle_manager_protocol.Conduit.receiver', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=926,
  serialized_end=969,
)


_PEERREQUEST = _descriptor.Descriptor(
  name='PeerRequest',
  full_name='muscle_manager_protocol.PeerRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='instance_name', full_name='muscle_manager_protocol.PeerRequest.instance_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=971,
  serialized_end=1007,
)


_PEERRESULT_PEERDIMENSIONS = _descriptor.Descriptor(
  name='PeerDimensions',
  full_name='muscle_manager_protocol.PeerResult.PeerDimensions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='peer_name', full_name='muscle_manager_protocol.PeerResult.PeerDimensions.peer_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dimensions', full_name='muscle_manager_protocol.PeerResult.PeerDimensions.dimensions', index=1,
      number=2, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1306,
  serialized_end=1361,
)

_PEERRESULT_PEERLOCATIONS = _descriptor.Descriptor(
  name='PeerLocations',
  full_name='muscle_manager_protocol.PeerResult.PeerLocations',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='instance_name', full_name='muscle_manager_protocol.PeerResult.PeerLocations.instance_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='locations', full_name='muscle_manager_protocol.PeerResult.PeerLocations.locations', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1363,
  serialized_end=1420,
)

_PEERRESULT = _descriptor.Descriptor(
  name='PeerResult',
  full_name='muscle_manager_protocol.PeerResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='muscle_manager_protocol.PeerResult.status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error_message', full_name='muscle_manager_protocol.PeerResult.error_message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='conduits', full_name='muscle_manager_protocol.PeerResult.conduits', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='peer_dimensions', full_name='muscle_manager_protocol.PeerResult.peer_dimensions', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='peer_locations', full_name='muscle_manager_protocol.PeerResult.peer_locations', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_PEERRESULT_PEERDIMENSIONS, _PEERRESULT_PEERLOCATIONS, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1010,
  serialized_end=1420,
)


_LISTOFDOUBLE = _descriptor.Descriptor(
  name='ListOfDouble',
  full_name='muscle_manager_protocol.ListOfDouble',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='values', full_name='muscle_manager_protocol.ListOfDouble.values', index=0,
      number=1, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1422,
  serialized_end=1452,
)


_LISTOFLISTOFDOUBLE = _descriptor.Descriptor(
  name='ListOfListOfDouble',
  full_name='muscle_manager_protocol.ListOfListOfDouble',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='values', full_name='muscle_manager_protocol.ListOfListOfDouble.values', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1454,
  serialized_end=1529,
)


_SETTING = _descriptor.Descriptor(
  name='Setting',
  full_name='muscle_manager_protocol.Setting',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='muscle_manager_protocol.Setting.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value_type', full_name='muscle_manager_protocol.Setting.value_type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value_string', full_name='muscle_manager_protocol.Setting.value_string', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value_int', full_name='muscle_manager_protocol.Setting.value_int', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value_float', full_name='muscle_manager_protocol.Setting.value_float', index=4,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value_bool', full_name='muscle_manager_protocol.Setting.value_bool', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value_list_float', full_name='muscle_manager_protocol.Setting.value_list_float', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value_list_list_float', full_name='muscle_manager_protocol.Setting.value_list_list_float', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1532,
  serialized_end=1841,
)


_SETTINGSREQUEST = _descriptor.Descriptor(
  name='SettingsRequest',
  full_name='muscle_manager_protocol.SettingsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1843,
  serialized_end=1860,
)


_SETTINGSRESULT = _descriptor.Descriptor(
  name='SettingsResult',
  full_name='muscle_manager_protocol.SettingsResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='setting_values', full_name='muscle_manager_protocol.SettingsResult.setting_values', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1862,
  serialized_end=1936,
)


_DEREGISTRATIONREQUEST = _descriptor.Descriptor(
  name='DeregistrationRequest',
  full_name='muscle_manager_protocol.DeregistrationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='instance_name', full_name='muscle_manager_protocol.DeregistrationRequest.instance_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1938,
  serialized_end=1984,
)


_DEREGISTRATIONRESULT = _descriptor.Descriptor(
  name='DeregistrationResult',
  full_name='muscle_manager_protocol.DeregistrationResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='muscle_manager_protocol.DeregistrationResult.status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error_message', full_name='muscle_manager_protocol.DeregistrationResult.error_message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1986,
  serialized_end=2086,
)

_LOGMESSAGE.fields_by_name['timestamp'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LOGMESSAGE.fields_by_name['level'].enum_type = _LOGLEVEL
_PORT.fields_by_name['operator'].enum_type = _OPERATOR
_PROFILEEVENT.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROFILEEVENT.fields_by_name['stop_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROFILEEVENT.fields_by_name['event_type'].enum_type = _PROFILEEVENTTYPE
_PROFILEEVENT.fields_by_name['port'].message_type = _PORT
_PROFILE.fields_by_name['events'].message_type = _PROFILEEVENT
_REGISTRATIONREQUEST.fields_by_name['ports'].message_type = _PORT
_REGISTRATIONRESULT.fields_by_name['status'].enum_type = _RESULTSTATUS
_PEERRESULT_PEERDIMENSIONS.containing_type = _PEERRESULT
_PEERRESULT_PEERLOCATIONS.containing_type = _PEERRESULT
_PEERRESULT.fields_by_name['status'].enum_type = _RESULTSTATUS
_PEERRESULT.fields_by_name['conduits'].message_type = _CONDUIT
_PEERRESULT.fields_by_name['peer_dimensions'].message_type = _PEERRESULT_PEERDIMENSIONS
_PEERRESULT.fields_by_name['peer_locations'].message_type = _PEERRESULT_PEERLOCATIONS
_LISTOFLISTOFDOUBLE.fields_by_name['values'].message_type = _LISTOFDOUBLE
_SETTING.fields_by_name['value_type'].enum_type = _SETTINGVALUETYPE
_SETTING.fields_by_name['value_list_float'].message_type = _LISTOFDOUBLE
_SETTING.fields_by_name['value_list_list_float'].message_type = _LISTOFLISTOFDOUBLE
_SETTINGSRESULT.fields_by_name['setting_values'].message_type = _SETTING
_DEREGISTRATIONRESULT.fields_by_name['status'].enum_type = _RESULTSTATUS
DESCRIPTOR.message_types_by_name['LogMessage'] = _LOGMESSAGE
DESCRIPTOR.message_types_by_name['LogResult'] = _LOGRESULT
DESCRIPTOR.message_types_by_name['Port'] = _PORT
DESCRIPTOR.message_types_by_name['ProfileEvent'] = _PROFILEEVENT
DESCRIPTOR.message_types_by_name['Profile'] = _PROFILE
DESCRIPTOR.message_types_by_name['ProfileResult'] = _PROFILERESULT
DESCRIPTOR.message_types_by_name['RegistrationRequest'] = _REGISTRATIONREQUEST
DESCRIPTOR.message_types_by_name['RegistrationResult'] = _REGISTRATIONRESULT
DESCRIPTOR.message_types_by_name['Conduit'] = _CONDUIT
DESCRIPTOR.message_types_by_name['PeerRequest'] = _PEERREQUEST
DESCRIPTOR.message_types_by_name['PeerResult'] = _PEERRESULT
DESCRIPTOR.message_types_by_name['ListOfDouble'] = _LISTOFDOUBLE
DESCRIPTOR.message_types_by_name['ListOfListOfDouble'] = _LISTOFLISTOFDOUBLE
DESCRIPTOR.message_types_by_name['Setting'] = _SETTING
DESCRIPTOR.message_types_by_name['SettingsRequest'] = _SETTINGSREQUEST
DESCRIPTOR.message_types_by_name['SettingsResult'] = _SETTINGSRESULT
DESCRIPTOR.message_types_by_name['DeregistrationRequest'] = _DEREGISTRATIONREQUEST
DESCRIPTOR.message_types_by_name['DeregistrationResult'] = _DEREGISTRATIONRESULT
DESCRIPTOR.enum_types_by_name['ResultStatus'] = _RESULTSTATUS
DESCRIPTOR.enum_types_by_name['Operator'] = _OPERATOR
DESCRIPTOR.enum_types_by_name['LogLevel'] = _LOGLEVEL
DESCRIPTOR.enum_types_by_name['ProfileEventType'] = _PROFILEEVENTTYPE
DESCRIPTOR.enum_types_by_name['SettingValueType'] = _SETTINGVALUETYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LogMessage = _reflection.GeneratedProtocolMessageType('LogMessage', (_message.Message,), dict(
  DESCRIPTOR = _LOGMESSAGE,
  __module__ = 'muscle_manager_protocol_pb2'
  # @@protoc_insertion_point(class_scope:muscle_manager_protocol.LogMessage)
  ))
_sym_db.RegisterMessage(LogMessage)

LogResult = _reflection.GeneratedProtocolMessageType('LogResult', (_message.Message,), dict(
  DESCRIPTOR = _LOGRESULT,
  __module__ = 'muscle_manager_protocol_pb2'
  # @@protoc_insertion_point(class_scope:muscle_manager_protocol.LogResult)
  ))
_sym_db.RegisterMessage(LogResult)

Port = _reflection.GeneratedProtocolMessageType('Port', (_message.Message,), dict(
  DESCRIPTOR = _PORT,
  __module__ = 'muscle_manager_protocol_pb2'
  # @@protoc_insertion_point(class_scope:muscle_manager_protocol.Port)
  ))
_sym_db.RegisterMessage(Port)

ProfileEvent = _reflection.GeneratedProtocolMessageType('ProfileEvent', (_message.Message,), dict(
  DESCRIPTOR = _PROFILEEVENT,
  __module__ = 'muscle_manager_protocol_pb2'
  # @@protoc_insertion_point(class_scope:muscle_manager_protocol.ProfileEvent)
  ))
_sym_db.RegisterMessage(ProfileEvent)

Profile = _reflection.GeneratedProtocolMessageType('Profile', (_message.Message,), dict(
  DESCRIPTOR = _PROFILE,
  __module__ = 'muscle_manager_protocol_pb2'
  # @@protoc_insertion_point(class_scope:muscle_manager_protocol.Profile)
  ))
_sym_db.RegisterMessage(Profile)

ProfileResult = _reflection.GeneratedProtocolMessageType('ProfileResult', (_message.Message,), dict(
  DESCRIPTOR = _PROFILERESULT,
  __module__ = 'muscle_manager_protocol_pb2'
  # @@protoc_insertion_point(class_scope:muscle_manager_protocol.ProfileResult)
  ))
_sym_db.RegisterMessage(ProfileResult)

RegistrationRequest = _reflection.GeneratedProtocolMessageType('RegistrationRequest', (_message.Message,), dict(
  DESCRIPTOR = _REGISTRATIONREQUEST,
  __module__ = 'muscle_manager_protocol_pb2'
  # @@protoc_insertion_point(class_scope:muscle_manager_protocol.RegistrationRequest)
  ))
_sym_db.RegisterMessage(RegistrationRequest)

RegistrationResult = _reflection.GeneratedProtocolMessageType('RegistrationResult', (_message.Message,), dict(
  DESCRIPTOR = _REGISTRATIONRESULT,
  __module__ = 'muscle_manager_protocol_pb2'
  # @@protoc_insertion_point(class_scope:muscle_manager_protocol.RegistrationResult)
  ))
_sym_db.RegisterMessage(RegistrationResult)

Conduit = _reflection.GeneratedProtocolMessageType('Conduit', (_message.Message,), dict(
  DESCRIPTOR = _CONDUIT,
  __module__ = 'muscle_manager_protocol_pb2'
  # @@protoc_insertion_point(class_scope:muscle_manager_protocol.Conduit)
  ))
_sym_db.RegisterMessage(Conduit)

PeerRequest = _reflection.GeneratedProtocolMessageType('PeerRequest', (_message.Message,), dict(
  DESCRIPTOR = _PEERREQUEST,
  __module__ = 'muscle_manager_protocol_pb2'
  # @@protoc_insertion_point(class_scope:muscle_manager_protocol.PeerRequest)
  ))
_sym_db.RegisterMessage(PeerRequest)

PeerResult = _reflection.GeneratedProtocolMessageType('PeerResult', (_message.Message,), dict(

  PeerDimensions = _reflection.GeneratedProtocolMessageType('PeerDimensions', (_message.Message,), dict(
    DESCRIPTOR = _PEERRESULT_PEERDIMENSIONS,
    __module__ = 'muscle_manager_protocol_pb2'
    # @@protoc_insertion_point(class_scope:muscle_manager_protocol.PeerResult.PeerDimensions)
    ))
  ,

  PeerLocations = _reflection.GeneratedProtocolMessageType('PeerLocations', (_message.Message,), dict(
    DESCRIPTOR = _PEERRESULT_PEERLOCATIONS,
    __module__ = 'muscle_manager_protocol_pb2'
    # @@protoc_insertion_point(class_scope:muscle_manager_protocol.PeerResult.PeerLocations)
    ))
  ,
  DESCRIPTOR = _PEERRESULT,
  __module__ = 'muscle_manager_protocol_pb2'
  # @@protoc_insertion_point(class_scope:muscle_manager_protocol.PeerResult)
  ))
_sym_db.RegisterMessage(PeerResult)
_sym_db.RegisterMessage(PeerResult.PeerDimensions)
_sym_db.RegisterMessage(PeerResult.PeerLocations)

ListOfDouble = _reflection.GeneratedProtocolMessageType('ListOfDouble', (_message.Message,), dict(
  DESCRIPTOR = _LISTOFDOUBLE,
  __module__ = 'muscle_manager_protocol_pb2'
  # @@protoc_insertion_point(class_scope:muscle_manager_protocol.ListOfDouble)
  ))
_sym_db.RegisterMessage(ListOfDouble)

ListOfListOfDouble = _reflection.GeneratedProtocolMessageType('ListOfListOfDouble', (_message.Message,), dict(
  DESCRIPTOR = _LISTOFLISTOFDOUBLE,
  __module__ = 'muscle_manager_protocol_pb2'
  # @@protoc_insertion_point(class_scope:muscle_manager_protocol.ListOfListOfDouble)
  ))
_sym_db.RegisterMessage(ListOfListOfDouble)

Setting = _reflection.GeneratedProtocolMessageType('Setting', (_message.Message,), dict(
  DESCRIPTOR = _SETTING,
  __module__ = 'muscle_manager_protocol_pb2'
  # @@protoc_insertion_point(class_scope:muscle_manager_protocol.Setting)
  ))
_sym_db.RegisterMessage(Setting)

SettingsRequest = _reflection.GeneratedProtocolMessageType('SettingsRequest', (_message.Message,), dict(
  DESCRIPTOR = _SETTINGSREQUEST,
  __module__ = 'muscle_manager_protocol_pb2'
  # @@protoc_insertion_point(class_scope:muscle_manager_protocol.SettingsRequest)
  ))
_sym_db.RegisterMessage(SettingsRequest)

SettingsResult = _reflection.GeneratedProtocolMessageType('SettingsResult', (_message.Message,), dict(
  DESCRIPTOR = _SETTINGSRESULT,
  __module__ = 'muscle_manager_protocol_pb2'
  # @@protoc_insertion_point(class_scope:muscle_manager_protocol.SettingsResult)
  ))
_sym_db.RegisterMessage(SettingsResult)

DeregistrationRequest = _reflection.GeneratedProtocolMessageType('DeregistrationRequest', (_message.Message,), dict(
  DESCRIPTOR = _DEREGISTRATIONREQUEST,
  __module__ = 'muscle_manager_protocol_pb2'
  # @@protoc_insertion_point(class_scope:muscle_manager_protocol.DeregistrationRequest)
  ))
_sym_db.RegisterMessage(DeregistrationRequest)

DeregistrationResult = _reflection.GeneratedProtocolMessageType('DeregistrationResult', (_message.Message,), dict(
  DESCRIPTOR = _DEREGISTRATIONRESULT,
  __module__ = 'muscle_manager_protocol_pb2'
  # @@protoc_insertion_point(class_scope:muscle_manager_protocol.DeregistrationResult)
  ))
_sym_db.RegisterMessage(DeregistrationResult)



_MUSCLEMANAGER = _descriptor.ServiceDescriptor(
  name='MuscleManager',
  full_name='muscle_manager_protocol.MuscleManager',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=2840,
  serialized_end=3466,
  methods=[
  _descriptor.MethodDescriptor(
    name='SubmitLogMessage',
    full_name='muscle_manager_protocol.MuscleManager.SubmitLogMessage',
    index=0,
    containing_service=None,
    input_type=_LOGMESSAGE,
    output_type=_LOGRESULT,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='SubmitProfileEvents',
    full_name='muscle_manager_protocol.MuscleManager.SubmitProfileEvents',
    index=1,
    containing_service=None,
    input_type=_PROFILE,
    output_type=_PROFILERESULT,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='RequestSettings',
    full_name='muscle_manager_protocol.MuscleManager.RequestSettings',
    index=2,
    containing_service=None,
    input_type=_SETTINGSREQUEST,
    output_type=_SETTINGSRESULT,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='RegisterInstance',
    full_name='muscle_manager_protocol.MuscleManager.RegisterInstance',
    index=3,
    containing_service=None,
    input_type=_REGISTRATIONREQUEST,
    output_type=_REGISTRATIONRESULT,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='RequestPeers',
    full_name='muscle_manager_protocol.MuscleManager.RequestPeers',
    index=4,
    containing_service=None,
    input_type=_PEERREQUEST,
    output_type=_PEERRESULT,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='DeregisterInstance',
    full_name='muscle_manager_protocol.MuscleManager.DeregisterInstance',
    index=5,
    containing_service=None,
    input_type=_DEREGISTRATIONREQUEST,
    output_type=_DEREGISTRATIONRESULT,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_MUSCLEMANAGER)

DESCRIPTOR.services_by_name['MuscleManager'] = _MUSCLEMANAGER

# @@protoc_insertion_point(module_scope)



from pathlib import Path
import sys

# Add libmuscle to the Python path, it seems pytest doesn't add it
# automatically if you're in a parallel directory like here.
sys.path.append(str(Path(__file__).parent / '..' / 'libmuscle' / 'python'))



import logging

from libmuscle.logging import LogLevel, Timestamp
from libmuscle.operator import Operator
from libmuscle.util import extract_log_file_location


class Logger:
    """The MUSCLE 3 Manager Logger component.

    The Logger component takes log messages and writes them to
    standard out.
    """
    def __init__(self) -> None:
        logfile = extract_log_file_location('muscle3_manager.log')
        self._local_handler = logging.FileHandler(str(logfile), mode='w')
        formatter = logging.Formatter('%(time_stamp)-15s: %(name)s'
                                      ' %(levelname)s: %(message)s')
        self._local_handler.setFormatter(formatter)

        # Find and remove default handler to disable automatic console output
        # Testing for 'stderr' in the stringified version is not nice, but
        # seems reliable, and doesn't mess up pytest's caplog mechanism while
        # it also doesn't introduce a runtime dependency on pytest.
        logging.getLogger().handlers = [
                h for h in logging.getLogger().handlers
                if 'stderr' not in str(h)]

        # add our own
        logging.getLogger().addHandler(self._local_handler)

        # hardwired for now
        logging.getLogger().setLevel(logging.INFO)
        logging.getLogger('yatiml.loader').setLevel(logging.WARNING)
        logging.getLogger('yatiml.dumper').setLevel(logging.WARNING)

    def close(self) -> None:
        logging.getLogger().removeHandler(self._local_handler)

    def log_message(
            self,
            instance_id: str,
            timestamp: Timestamp,
            level: LogLevel,
            text: str) -> None:
        """Log a message.

        Args:
            instance_id: Identifier of the instance that generated the \
                    message.
            timestamp: Time when this log message was generated, \
                    according to the clock on that machine.
            level: The log level of the message.
            text: The message text.
        """
        logger = logging.getLogger(instance_id)
        logger.log(
                level.as_python_level(),
                text,
                extra={'time_stamp': timestamp.to_rfc3339()})



from ymmsl import Reference


class Client:
    """A client that connects to an MCP server.

    This is a base class for MCP Clients. An MCP Client connects to
    an MCP Server over some lower-level communication protocol, and
    requests messages from it.
    """
    @staticmethod
    def can_connect_to(location: str) -> bool:
        """Whether this client class can connect to the given location.

        Args:
            location: The location to potentially connect to.

        Returns:
            True iff this class can connect to this location.
        """
        raise NotImplementedError()     # pragma: no cover

    @staticmethod
    def shutdown(instance_id: Reference) -> None:
        """Shut down and free any resources shared by all clients.

        This is an optional hook for communication subsystems that
        need it. If implemented, it must work correctly even if no
        clients have ever been instantiated.

        This will be called after all clients of this class have been
        closed.
        """
        pass

    def __init__(self, instance_id: Reference, location: str) -> None:
        """Create an MCPClient for a given location.

        The client will connect to this location and be able to request
        messages from any compute element and port represented by it.

        Args:
            instance_id: Id of our instance.
            location: A location string for the peer.
        """
        self._location = location

    def receive(self, receiver: Reference) -> bytes:
        """Receive a message from a port this client connects to.

        Args:
            receiver: The receiving (local) port.

        Returns:
            The received message.
        """
        raise NotImplementedError()     # pragma: no cover

    def close(self) -> None:
        """Closes this client.

        This closes any connections this client has and/or performs
        other shutdown activities.
        """
        raise NotImplementedError()     # pragma: no cover



import pytest
from ymmsl import Reference

from libmuscle.mcp.direct_server import DirectServer
from libmuscle.mcp.tcp_server import TcpServer
from libmuscle.mcp.message import Message
from libmuscle.outbox import Outbox
from libmuscle.post_office import PostOffice


@pytest.fixture
def receiver():
    return Reference('test_receiver.test_port')


@pytest.fixture
def post_office(receiver):
    class MockPO(PostOffice):
        outboxes = {receiver: Outbox()}

        def get_message(self, receiver: Reference) -> Message:
            return self.outboxes[receiver].retrieve()

    return MockPO()


@pytest.fixture
def direct_server(post_office):
    return DirectServer('test_sender', post_office)


@pytest.fixture
def tcp_server(post_office):
    server = TcpServer('test_sender', post_office)
    yield server
    server.close()



from collections import OrderedDict
import sys
from typing import List

import pytest
from ymmsl import (ComputeElement, Conduit, Configuration, Model, Operator,
                   Reference, Settings)

from libmuscle import Instance, Message
from libmuscle.runner import run_simulation


def macro():
    """Macro model implementation.
    """
    instance = Instance({
            Operator.O_I: ['out[]'],
            Operator.S: ['in[]']})

    while instance.reuse_instance():
        # f_init
        assert instance.get_setting('test1') == 13

        # o_i
        assert instance.is_vector_port('out')
        for slot in range(10):
            instance.send('out', Message(0.0, 10.0, 'testing'), slot)

        # s/b
        for slot in range(10):
            msg = instance.receive('in', slot)
            assert msg.data == 'testing back'


def micro():
    """Micro model implementation.
    """
    instance = Instance({
            Operator.F_INIT: ['in'],
            Operator.O_F: ['out']})

    while instance.reuse_instance():
        # f_init
        assert instance.get_setting('test3', 'str') == 'testing'
        assert instance.get_setting('test4', 'bool') is True
        assert instance.get_setting('test6', '[[float]]')[0][1] == 2.0

        msg = instance.receive('in')
        assert msg.data == 'testing'

        # o_f
        instance.send('out', Message(0.1, None, 'testing back'))


def test_all(log_file_in_tmpdir):
    """A positive all-up test of everything.
    """
    elements = [
            ComputeElement('macro', 'macro_impl'),
            ComputeElement('micro', 'micro_impl', [10])]

    conduits = [
            Conduit('macro.out', 'micro.in'),
            Conduit('micro.out', 'macro.in')]

    model = Model('test_model', elements, conduits)
    settings = Settings(OrderedDict([
                ('test1', 13),
                ('test2', 13.3),
                ('test3', 'testing'),
                ('test4', True),
                ('test5', [2.3, 5.6]),
                ('test6', [[1.0, 2.0], [3.0, 1.0]])]))

    configuration = Configuration(model, settings)

    implementations = {'macro_impl': macro, 'micro_impl': micro}
    run_simulation(configuration, implementations)



import pytest
from ymmsl import (ComputeElement, Conduit, Configuration, Identifier, Model,
                   Reference, Settings)

from libmuscle.manager.instance_registry import InstanceRegistry
from libmuscle.manager.logger import Logger
from libmuscle.manager.mmp_server import MMPServicer
from libmuscle.manager.topology_store import TopologyStore


@pytest.fixture
def logger():
    test_logger = Logger()
    yield test_logger
    test_logger.close()


@pytest.fixture
def settings():
    return Settings()


@pytest.fixture
def instance_registry():
    expected_instances = ['macro']
    for i in range(10):
        expected_instances.append('micro[{}]'.format(i))
    return InstanceRegistry(expected_instances)


@pytest.fixture
def topology_store() -> TopologyStore:
    config = Configuration(
            Model(
                'test_model',
                [
                    ComputeElement('macro', 'macro_implementation'),
                    ComputeElement(
                        'micro', 'micro_implementation', [10, 10])],
                [
                    Conduit('macro.out', 'micro.in'),
                    Conduit('micro.out', 'macro.in')
                ]))

    return TopologyStore(config)


@pytest.fixture
def mmp_servicer(logger, settings, instance_registry, topology_store):
    return MMPServicer(logger, settings, instance_registry,
                       topology_store)


@pytest.fixture
def loaded_instance_registry(instance_registry):
    instance_registry.add(Reference('macro'), ['direct:macro'], [])
    for j in range(10):
        for i in range(10):
            name = Reference('micro') + j + i
            location = 'direct:{}'.format(name)
            instance_registry.add(name, [location], [])
    return instance_registry


@pytest.fixture
def registered_mmp_servicer(logger, settings, loaded_instance_registry,
                            topology_store):
    return MMPServicer(logger, settings, loaded_instance_registry,
                       topology_store)


@pytest.fixture
def topology_store2() -> TopologyStore:
    config = Configuration(
            Model(
                'test_model',
                [
                    ComputeElement('macro', 'macro_implementation'),
                    ComputeElement('meso', 'meso_implementation', [5]),
                    ComputeElement('micro', 'micro_implementation', [5, 10])
                ],
                [
                    Conduit('macro.out', 'meso.in'),
                    Conduit('meso.out', 'micro.in'),
                    Conduit('micro.out', 'meso.in'),
                    Conduit('meso.out', 'macro.in')
                ]))

    return TopologyStore(config)


@pytest.fixture
def loaded_instance_registry2():
    expected_instances = ['macro']
    for i in range(5):
        expected_instances.append('meso[{}]'.format(i))
    for j in range(5):
        for i in range(10):
            expected_instances.append('micro[{}][{}]'.format(j, i))
    instance_registry = InstanceRegistry(expected_instances)

    instance_registry.add(Reference('macro'), ['direct:macro'], [])

    for j in range(5):
        name = Reference('meso') + j
        location = 'direct:{}'.format(name)
        instance_registry.add(name, [location], [])

    for j in range(5):
        for i in range(10):
            name = Reference('micro') + j + i
            location = 'direct:{}'.format(name)
            instance_registry.add(name, [location], [])
    return instance_registry


@pytest.fixture
def registered_mmp_servicer2(
        logger, settings, loaded_instance_registry2, topology_store2):
    return MMPServicer(logger, settings, loaded_instance_registry2,
                       topology_store2)



from typing import List, Optional
from ymmsl import Identifier, Operator, Reference
import ymmsl

from libmuscle.operator import operator_from_grpc, operator_to_grpc
import muscle_manager_protocol.muscle_manager_protocol_pb2 as mmp


# Convert between grpc and ymmsl Port types
def port_from_grpc(port: mmp.Port) -> ymmsl.Port:
    return ymmsl.Port(
            Identifier(port.name),
            operator_from_grpc(port.operator))


def port_to_grpc(port: ymmsl.Port) -> mmp.Port:
    return mmp.Port(
            name=str(port.name),
            operator=operator_to_grpc(port.operator))


def optional_port_to_grpc(port: Optional[ymmsl.Port]) -> Optional[mmp.Port]:
    if port is None:
        return None
    return port_to_grpc(port)


class Port(ymmsl.Port):
    """Represents a gateway to the outside world.

    Ports can be used to send or receive messages. They have a name and
    an operator, as well as a set of dimensions that determines the
    valid slot indices for sending or receiving on this port.

    Attributes:
        name (Identifier): Name of this port.
        operator (Operator): Operator associated with this port.
    """
    def __init__(self, name: str, operator: Operator, is_vector: bool,
                 is_connected: bool, our_ndims: int, peer_dims: List[int]
                 ) -> None:
        """Create a Port.

        Args:
            name: Name of this port.
            operator: Corresponding operator.
            is_vector: Whether this is a vector port.
            is_connected: Whether this port is connected to a peer.
            our_ndims: Number of dimensions of our instance set.
            peer_dims: Dimensions of the peer instance set of this port.
        """
        super().__init__(Identifier(name), operator)

        self._is_connected = is_connected

        if is_vector:
            if our_ndims == len(peer_dims):
                self._length = 0    # type: Optional[int]
            elif our_ndims + 1 == len(peer_dims):
                self._length = peer_dims[-1]
            elif our_ndims > len(peer_dims):
                raise RuntimeError(('Vector port "{}" is connected to an'
                                    ' instance set with fewer dimensions.'
                                    ' It should be connected to a scalar'
                                    ' port on a set with one more dimension,'
                                    ' or to a vector port on a set with the'
                                    ' same number of dimensions.').format(
                                        name))
            else:
                raise RuntimeError(('Port "{}" is connected to an instance set'
                                    ' with more than one dimension more than'
                                    ' its own, which is not possible.').format(
                                        name))
            self._is_open = [True] * self._length
        else:
            if our_ndims < len(peer_dims):
                raise RuntimeError(('Scalar port "{}" is connected to an'
                                    ' instance set with more dimensions.'
                                    ' It should be connected to a scalar'
                                    ' port on an instance set with the same'
                                    ' dimensions, or to a vector port on an'
                                    ' instance set with one less dimension.'
                                    ).format(name))
            elif our_ndims > len(peer_dims) + 1:
                raise RuntimeError(('Scalar port "{}" is connected to an'
                                    ' instance set with at least two fewer'
                                    ' dimensions, which is not possible.'
                                    ).format(name))
            self._length = None
            self._is_open = [True]

        self._is_resizable = is_vector and (our_ndims == len(peer_dims))

    # Note: I'm not sure how this will develop exactly, so this class has some
    # accessors even if those are un-Pythonic; in the future a simple variable
    # read may not be the right model.

    def is_connected(self) -> bool:
        """Returns whether the port is connected to a peer.

        Returns:
            True if there is a peer, False if there is not.
        """
        return self._is_connected

    def is_open(self, slot: Optional[int] = None) -> bool:
        """Returns whether this port is open.
        """
        if slot is not None:
            return self._is_open[slot]
        return self._is_open[0]

    def is_vector(self) -> bool:
        """Returns whether this is a vector port.

        Returns:
            True if it is vector, False if it is scalar.
        """
        return self._length is not None

    def is_resizable(self) -> bool:
        """Returns whether this port can be resized.
        """
        return self._is_resizable

    def get_length(self) -> int:
        """Returns the length of this port.

        Raises:
            RuntimeError: If this port is a scalar port.
        """
        if self._length is None:
            raise RuntimeError(('Tried to get length of scalar port {}'
                                ).format(self.name))
        return self._length

    def set_length(self, length: int) -> None:
        """Sets the length of a resizable vector port.

        Only call this if is_resizable() returns True.

        Args:
            length: The new length.

        Raises:
            RuntimeError: If the port is not resizable.
        """
        if not self._is_resizable:
            raise RuntimeError(('Tried to resize port {}, but it is not'
                                ' resizable.'.format(self.name)))
        if length != self._length:
            self._length = length
            self._is_open = [True] * self._length

    def set_closed(self, slot: Optional[int] = None) -> None:
        """Marks this port as closed.
        """
        if slot is not None:
            self._is_open[slot] = False
        else:
            self._is_open = [False]



import logging
import multiprocessing as mp
import sys
from typing import Generator

import pytest
import yatiml
import ymmsl

import integration_test.include_libmuscle

from libmuscle.manager.instance_registry import InstanceRegistry
from libmuscle.manager.logger import Logger
from libmuscle.manager.mmp_server import MMPServer
from libmuscle.manager.manager import elements_for_model
from libmuscle.manager.topology_store import TopologyStore


@pytest.fixture
def yatiml_log_warning():
    yatiml.logger.setLevel(logging.WARNING)


def make_server(ymmsl_doc: ymmsl.Configuration):
    logger = Logger()
    expected_elements = elements_for_model(ymmsl_doc.model)
    instance_registry = InstanceRegistry(expected_elements)
    topology_store = TopologyStore(ymmsl_doc)
    server = MMPServer(logger, ymmsl_doc.settings, instance_registry,
                       topology_store)
    return server


def start_mmp_server(control_pipe, ymmsl_doc):
    control_pipe[0].close()
    server = make_server(ymmsl_doc)
    control_pipe[1].send(True)
    control_pipe[1].recv()
    control_pipe[1].close()
    server.stop()


def make_server_process(ymmsl_doc):
    control_pipe = mp.Pipe()
    process = mp.Process(target=start_mmp_server,
                         args=(control_pipe, ymmsl_doc),
                         name='MMPServer')
    process.start()
    control_pipe[1].close()
    # wait for start
    control_pipe[0].recv()
    yield None
    control_pipe[0].send(True)
    control_pipe[0].close()
    process.join()


@pytest.fixture
def mmp_server_process(yatiml_log_warning):
    ymmsl_text = (
            'ymmsl_version: v0.1\n'
            'model:\n'
            '  name: test_model\n'
            '  compute_elements:\n'
            '    macro: macro_implementation\n'
            '    micro:\n'
            '      implementation: micro_implementation\n'
            '      multiplicity: [10]\n'
            '  conduits:\n'
            '    macro.out: micro.in\n'
            '    micro.out: macro.in\n'
            'settings:\n'
            '  test1: 13\n'
            '  test2: 13.3\n'
            '  test3: testing\n'
            '  test4: True\n'
            '  test5: [2.3, 5.6]\n'
            '  test6:\n'
            '    - [1.0, 2.0]\n'
            '    - [3.0, 1.0]\n'
            )
    ymmsl_doc = ymmsl.load(ymmsl_text)

    yield from make_server_process(ymmsl_doc)


@pytest.fixture
def mmp_server_process_simple(yatiml_log_warning):
    ymmsl_text = (
            'ymmsl_version: v0.1\n'
            'model:\n'
            '  name: test_model\n'
            '  compute_elements:\n'
            '    macro: macro_implementation\n'
            '    micro: micro_implementation\n'
            '  conduits:\n'
            '    macro.out: micro.in\n'
            '    micro.out: macro.in\n'
            'settings:\n'
            '  test1: 13\n'
            '  test2: 13.3\n'
            '  test3: testing\n'
            '  test4: True\n'
            '  test5: [2.3, 5.6]\n'
            '  test6:\n'
            '    - [1.0, 2.0]\n'
            '    - [3.0, 1.0]\n'
            )
    ymmsl_doc = ymmsl.load(ymmsl_text)

    yield from make_server_process(ymmsl_doc)


@pytest.fixture
def mmp_server(yatiml_log_warning):
    ymmsl_text = (
            'ymmsl_version: v0.1\n'
            'model:\n'
            '  name: test_model\n'
            '  compute_elements:\n'
            '    macro: macro_implementation\n'
            '    micro:\n'
            '      implementation: micro_implementation\n'
            '      multiplicity: [10]\n'
            '  conduits:\n'
            '    macro.out: micro.in\n'
            '    micro.out: macro.in\n'
            'settings:\n'
            '  test1: 13\n'
            '  test2: 13.3\n'
            '  test3: testing\n'
            '  test4: True\n'
            '  test5: [2.3, 5.6]\n'
            '  test6:\n'
            '    - [1.0, 2.0]\n'
            '    - [3.0, 1.0]\n'
            )
    ymmsl_doc = ymmsl.load(ymmsl_text)

    server = make_server(ymmsl_doc)
    yield server
    server.stop()


@pytest.fixture
def sys_argv_manager() -> Generator[None, None, None]:
    old_argv = sys.argv
    sys.argv = sys.argv + ['--muscle-manager=localhost:9000']
    yield None
    sys.argv = old_argv


@pytest.fixture
def log_file_in_tmpdir(tmpdir):
    old_argv = sys.argv
    sys.argv = sys.argv + ['--muscle-log-file={}'.format(tmpdir)]
    yield None
    sys.argv = old_argv



import time
from ymmsl import Reference

import pytest

from libmuscle.mcp.tcp_client import TcpClient
from libmuscle.mcp.tcp_server import TcpServer
from libmuscle.mcp.message import Message


def test_send_receive(receiver, post_office):
    message = b'testing'

    # prepare post office
    post_office.outboxes[receiver].deposit(message)

    # create server
    sender_instance_id = Reference('test_sender')
    server = TcpServer(sender_instance_id, post_office)

    # create client
    recv_instance_id = Reference('test_receiver')

    server_location = server.get_location()
    assert TcpClient.can_connect_to(server_location)
    client = TcpClient(recv_instance_id, server_location)

    message2 = client.receive(receiver)
    assert message == message2

    client.close()
    TcpClient.shutdown(recv_instance_id)

    server.close()



from enum import Enum
from threading import Condition
from typing import Dict  # noqa
from typing import cast, List

from ymmsl import Port, Reference


class _InstanceStatus(Enum):
    EXPECTED = 0,
    REGISTERED = 1,
    DEREGISTERED = 2


class InstanceRegistry:
    """Keeps track of running instances.

    The InstanceRegistry is a simple in-memory database that stores
    information about running instances of compute elements.
    """
    def __init__(self, expected_instances: List[str]) -> None:
        """Construct an empty InstanceRegistry.

        Args:
            expected_instances: List of instance names expected to
                    register.
        """
        self.__status = dict()  # type: Dict[Reference, _InstanceStatus]
        self.__deregistered_one = Condition()   # doubles as lock for __status
        self.__locations = dict()  # type: Dict[Reference, List[str]]
        self.__ports = dict()  # type: Dict[Reference, List[Port]]

        for instance_name in expected_instances:
            self.__status[Reference(instance_name)] = _InstanceStatus.EXPECTED

    def add(self, name: Reference, locations: List[str], ports: List[Port]
            ) -> None:
        """Add an instance to the registry.

        Args:
            name: Name of the instance.
            locations: Network locations where it can be reached.
            ports: List of ports of this instance.

        Raises:
            ValueError: If an instance with this name has already been
                    registered.
        """
        if name in self.__locations or name in self.__ports:
            raise ValueError('Instance already registered')
        with self.__deregistered_one:
            self.__status[name] = _InstanceStatus.REGISTERED
        self.__locations[name] = locations
        self.__ports[name] = ports

    def get_locations(self, name: Reference) -> List[str]:
        """Retrieves the locations of a registered instance.

        Args:
            name: The name of the instance to get the location of.

        Raises:
            KeyError: If no instance with this name was registered.
        """
        return self.__locations[name]

    def get_ports(self, name: Reference) -> List[Port]:
        """Retrieves the ports of a registered instance.

        Args:
            name: The name of the instance whose ports to retrieve.

        Raises:
            KeyError: If no instance with this name was registered.
        """
        return self.__ports[name]

    def remove(self, name: Reference) -> None:
        """Remove an instance from the registry.

        Args:
            name: Name of the instance to remove.

        Raises:
            KeyError: If the instance does not exist.
        """
        del(self.__locations[name])
        del(self.__ports[name])
        with self.__deregistered_one:
            self.__status[name] = _InstanceStatus.DEREGISTERED
            self.__deregistered_one.notify()

    def wait(self) -> None:
        """Waits until all instance are deregistered.

        This function blocks, and returns after each expected instance
        has been registered and deregistered again, signalling the end
        of the simulation run.
        """
        # this is called from a different thread than add/remove
        def all_deregistered() -> bool:
            return all(map(
                    lambda x: x == _InstanceStatus.DEREGISTERED,
                    self.__status.values()))

        with self.__deregistered_one:
            while not all_deregistered():
                self.__deregistered_one.wait()



from typing import Dict

from ymmsl import Reference

from libmuscle.post_office import PostOffice


class ServerNotSupported(RuntimeError):
    pass


class Server:
    """A server that accepts MCP connections.

    This is a base class for MCP Servers. An MCP Server accepts
    connections over some lower-level communication protocol, and
    processes message requests by sending the requested message.
    """
    def __init__(self, instance_id: Reference, post_office: PostOffice
                 ) -> None:
        """Create a Server.

        Args:
            instance_id: Id of the instance we're a server for.
            post_office: A PostOffice to obtain data from.
        """
        self._instance_id = instance_id
        self._post_office = post_office

    def get_location(self) -> str:
        """Returns the location this server listens on.

        Returns:
            A string containing the location.
        """
        raise NotImplementedError()  # pragma: no cover

    def close(self) -> None:
        """Closes this server.

        Stops the server listening, waits for existing clients to
        disconnect, then frees any other resources.
        """
        raise NotImplementedError()  # pragma: no cover



import click
import ymmsl

from libmuscle.manager.manager import start_server


@click.command()
@click.argument('ymmsl_file')
def manage_simulation(ymmsl_file: str) -> None:
    with open(ymmsl_file) as f:
        configuration = ymmsl.load(f)

    server = start_server(configuration)
    print(server.get_location())
    server.wait()


if __name__ == '__main__':
    manage_simulation()



from typing import Dict

from ymmsl import Operator
import muscle_manager_protocol.muscle_manager_protocol_pb2 as mmp


def operator_from_grpc(
        operator: mmp.Operator
        ) -> Operator:
    """Creates an operator from a gRPC-generated Operator.

    Args:
        operator: An operator, received from gRPC.

    Returns:
        The same operator, as an Operator.
    """
    operator_map = {
            mmp.OPERATOR_NONE: Operator.NONE,
            mmp.OPERATOR_F_INIT: Operator.F_INIT,
            mmp.OPERATOR_O_I: Operator.O_I,
            mmp.OPERATOR_S: Operator.S,
            mmp.OPERATOR_B: Operator.B,
            mmp.OPERATOR_O_F: Operator.O_F
            }   # type: Dict[int, Operator]
    return operator_map[operator]


def operator_to_grpc(operator: Operator) -> mmp.Operator:
    """Converts the operator to the gRPC generated type.

    Returns:
        The operator, as the gRPC type.
    """
    operator_map = {
            Operator.NONE: mmp.OPERATOR_NONE,
            Operator.F_INIT: mmp.OPERATOR_F_INIT,
            Operator.O_I: mmp.OPERATOR_O_I,
            Operator.S: mmp.OPERATOR_S,
            Operator.B: mmp.OPERATOR_B,
            Operator.O_F: mmp.OPERATOR_O_F
            }   # type: Dict[Operator, mmp.Operator]
    return operator_map[operator]



from random import uniform
from time import perf_counter, sleep
from typing import Dict, Iterable, List, Tuple

import grpc
from ymmsl import Conduit, Port, Reference, Settings

import muscle_manager_protocol.muscle_manager_protocol_pb2 as mmp
import muscle_manager_protocol.muscle_manager_protocol_pb2_grpc as mmp_grpc

from libmuscle.port import port_to_grpc
from libmuscle.profiling import ProfileEvent
from libmuscle.logging import LogMessage


CONNECTION_TIMEOUT = 300
PEER_TIMEOUT = 600
PEER_INTERVAL_MIN = 5.0
PEER_INTERVAL_MAX = 10.0


def conduit_from_grpc(conduit: mmp.Conduit) -> Conduit:
    """Converts a Conduit from grpc to ymmsl.

    Args:
        conduit: A conduit.

    Returns:
        The same conduit, as ymmsl.Conduit.
    """
    return Conduit(conduit.sender, conduit.receiver)


class MMPClient():
    """The client for the MUSCLE Manager Protocol.

    This class connects to the Manager and communicates with it on \
    behalf of the rest of libmuscle.

    It manages the connection, and converts between our native types \
    and the gRPC generated types.
    """
    def __init__(self, location: str) -> None:
        """Create an MMPClient.

        Args:
            location: A connection string of the form hostname:port
        """
        channel = grpc.insecure_channel(location)
        ready = grpc.channel_ready_future(channel)
        try:
            ready.result(timeout=CONNECTION_TIMEOUT)
        except grpc.FutureTimeoutError:
            ready.cancel()
            channel.close()
            raise RuntimeError('Failed to connect to the MUSCLE manager')

        self.__client = mmp_grpc.MuscleManagerStub(channel)

    def submit_log_message(self, message: LogMessage) -> None:
        """Send a log message to the manager.

        Args:
            message: The message to send.
        """
        self.__client.SubmitLogMessage(message.to_grpc())

    def submit_profile_events(self, events: Iterable[ProfileEvent]) -> None:
        """Sends profiling events to the manager.

        Args:
            events: The events to send.
        """
        self.__client.SubmitProfileEvents(mmp.Profile(events=[
            e.to_grpc() for e in events]))

    def get_settings(self) -> Settings:
        """Get the central settings from the manager.

        Returns:
            The requested settings.
        """
        LLF = mmp.SETTING_VALUE_TYPE_LIST_LIST_FLOAT
        result = self.__client.RequestSettings(mmp.SettingsRequest())
        settings = Settings()
        for setting in result.setting_values:
            if setting.value_type == mmp.SETTING_VALUE_TYPE_STRING:
                settings[setting.name] = setting.value_string
            elif setting.value_type == mmp.SETTING_VALUE_TYPE_INT:
                settings[setting.name] = setting.value_int
            elif setting.value_type == mmp.SETTING_VALUE_TYPE_FLOAT:
                settings[setting.name] = setting.value_float
            elif setting.value_type == mmp.SETTING_VALUE_TYPE_BOOL:
                settings[setting.name] = setting.value_bool
            elif setting.value_type == mmp.SETTING_VALUE_TYPE_LIST_FLOAT:
                settings[setting.name] = list(
                        setting.value_list_float.values)
            elif setting.value_type == LLF:
                rows = list()   # type: List[List[float]]
                for mmp_row in setting.value_list_list_float.values:
                    rows.append(list(mmp_row.values))
                settings[setting.name] = rows
        return settings

    def register_instance(self, name: Reference, locations: List[str],
                          ports: List[Port]) -> None:
        """Register a compute element instance with the manager.

        Args:
            name: Name of the instance in the simulation.
            locations: List of places where the instance can be
                    reached.
            ports: List of ports of this instance.
        """
        grpc_ports = map(port_to_grpc, ports)
        request = mmp.RegistrationRequest(
                instance_name=str(name),
                network_locations=locations,
                ports=grpc_ports)
        self.__client.RegisterInstance(request)

    def request_peers(self, name: Reference
                      ) -> Tuple[List[Conduit],
                                 Dict[Reference, List[int]],
                                 Dict[Reference, List[str]]]:
        """Request connection information about peers.

        This will repeat the request at an exponentially increasing
        query interval at first, until it reaches the interval
        specified by PEER_INTERVAL_MIN and PEER_INTERVAL_MAX. From
        there on, intervals are drawn randomly from that range.

        Args:
            name: Name of the current instance.

        Returns:
            A tuple containing a list of conduits that this instance is
            attached to, a dictionary of peer dimensions, which is
            indexed by Reference to the peer kernel, and specifies how
            many instances of that kernel there are, and a dictionary
            of peer instance locations, indexed by Reference to a peer
            instance, and containing for each peer instance a list of
            network location strings at which it can be reached.
        """
        sleep_time = 0.1
        start_time = perf_counter()
        request = mmp.PeerRequest(instance_name=str(name))
        result = self.__client.RequestPeers(request)
        while (result.status == mmp.RESULT_STATUS_PENDING and
               perf_counter() < start_time + PEER_TIMEOUT and
               sleep_time < PEER_INTERVAL_MIN):
            sleep(sleep_time)
            result = self.__client.RequestPeers(request)
            sleep_time *= 1.5

        while (result.status == mmp.RESULT_STATUS_PENDING and
               perf_counter() < start_time + PEER_TIMEOUT):
            sleep(uniform(PEER_INTERVAL_MIN, PEER_INTERVAL_MAX))
            result = self.__client.RequestPeers(request)

        if result.status == mmp.RESULT_STATUS_PENDING:
            raise RuntimeError('Timeout waiting for peers to appear')

        if result.status == mmp.RESULT_STATUS_ERROR:
            raise RuntimeError('Error getting peers from manager: {}'.format(
                    result.error_message))

        conduits = list(map(conduit_from_grpc, result.conduits))

        peer_dimensions = dict()
        for peer_dimension in result.peer_dimensions:
            peer_dimensions[peer_dimension.peer_name] = \
                    peer_dimension.dimensions

        peer_locations = dict()
        for peer_location in result.peer_locations:
            peer_locations[peer_location.instance_name] = \
                    peer_location.locations

        return conduits, peer_dimensions, peer_locations

    def deregister_instance(self, name: Reference) -> None:
        """Deregister a compute element instance with the manager.

        Args:
            name: Name of the instance in the simulation.
        """
        request = mmp.DeregistrationRequest(
                instance_name=str(name))
        self.__client.DeregisterInstance(request)



from libmuscle.mcp.direct_client import DirectClient
from libmuscle.mcp.direct_server import DirectServer
from libmuscle.mcp.pipe_client import PipeClient
from libmuscle.mcp.pipe_server import PipeServer
from libmuscle.mcp.tcp_client import TcpClient
from libmuscle.mcp.tcp_server import TcpServer


# These must be in order of preference, i.e. most efficient first
client_types = [DirectClient, PipeClient, TcpClient]


server_types = [DirectServer, PipeServer, TcpServer]



from typing import Dict, List

from ymmsl import Conduit, Configuration, Model, Reference


class TopologyStore:
    """Holds a description of how the simulation is wired together.

    This class contains the list of conduits through which the
    submodels are connected.

    Attributes:
        conduits (List[Conduit]): A list of conduits.
    """
    def __init__(self, config: Configuration) -> None:
        """Creates a TopologyStore.

        Creates a TopologyStore containing conduits read from the given
        configuration data, which must contain a 'model' key.

        Args:
            configuration: A yMMSL configuration.
        """
        if config.model is None or not isinstance(config.model, Model):
            raise ValueError('The yMMSL experiment description does not'
                             ' contain a (complete) model section, so there'
                             ' is nothing to run!')
        self.conduits = config.model.conduits
        self.kernel_dimensions = {
                k.name: k.multiplicity
                for k in config.model.compute_elements}

    def has_kernel(self, kernel: Reference) -> bool:
        """Returns True iff the given kernel is in the model.

        Args:
            kernel: The kernel to check for.
        """
        return kernel in self.kernel_dimensions

    def get_conduits(self, kernel_name: Reference) -> List[Conduit]:
        """Returns the list of conduits that attach to the given kernel.

        Args:
            kernel_name: Name of the kernel.

        Returns:
            All conduits that this kernel is a sender or receiver of.
        """
        ret = list()
        for conduit in self.conduits:
            if conduit.sending_compute_element() == kernel_name:
                ret.append(conduit)
            if conduit.receiving_compute_element() == kernel_name:
                ret.append(conduit)
        return ret

    def get_peer_dimensions(self, kernel_name: Reference
                            ) -> Dict[Reference, List[int]]:
        """Returns the dimensions of peer kernels.

        For each kernel that the given kernel shares a conduit with,
        the returned dictionary has an entry containing its dimensions.

        Args:
            kernel_name: Name of the kernel for which to get peers.

        Returns:
            A dict of peer kernels and their dimensions.
        """
        ret = dict()
        for conduit in self.conduits:
            if conduit.sending_compute_element() == kernel_name:
                recv = conduit.receiving_compute_element()
                ret[recv] = self.kernel_dimensions[recv]
            if conduit.receiving_compute_element() == kernel_name:
                snd = conduit.sending_compute_element()
                ret[snd] = self.kernel_dimensions[snd]
        return ret



import os
import sys


# The gRPC generated code contains an absolute import. So either it
# needs to be in the top-level directory, or the user needs to modify
# their PYTHONPATH environment variable, or we add it to the path here.
sys.path.append(os.path.dirname(os.path.expanduser(__file__)))



from typing import List

import pytest
from ymmsl import (ComputeElement, Conduit, Configuration, Operator, Reference,
                   Model, Settings)

from libmuscle import Instance, Message
from libmuscle.runner import run_simulation


def duplication_mapper():
    """Duplication mapper implementation.
    """
    instance = Instance()

    while instance.reuse_instance():
        # o_f
        out_ports = instance.list_ports()[Operator.O_F]

        message = Message(0.0, None, 'testing')
        for out_port in out_ports:
            instance.send(out_port, message)


def receiver():
    """Receiver for messages from dm.
    """
    instance = Instance({Operator.F_INIT: ['in']})

    while instance.reuse_instance():
        # f_init
        msg = instance.receive('in')
        assert msg.data == 'testing'


def test_duplication_mapper(log_file_in_tmpdir):
    """A positive all-up test of duplication mappers.

    This is an acyclic workflow.
    """
    elements = [
            ComputeElement('dm', 'muscle.duplication_mapper'),
            ComputeElement('first', 'receiver'),
            ComputeElement('second', 'receiver')]

    conduits = [
                Conduit('dm.out', 'first.in'),
                Conduit('dm.out2', 'second.in')]

    model = Model('test_model', elements, conduits)
    settings = Settings()

    configuration = Configuration(model, settings)

    implementations = {
            'muscle.duplication_mapper': duplication_mapper,
            'receiver': receiver}
    run_simulation(configuration, implementations)



from time import time
from typing import List, Optional

from ymmsl import Port, Reference

import muscle_manager_protocol.muscle_manager_protocol_pb2 as mmp
from libmuscle.mmp_client import MMPClient
from libmuscle.profiling import ProfileEvent, ProfileEventType
from libmuscle.timestamp import Timestamp


class Profiler:
    """Collects profiling events and sends them to the manager.
    """
    def __init__(self, instance_id: Reference, manager: MMPClient) -> None:
        """Create a Profiler.

        Args:
            manager: The client used to submit data to the manager.
        """
        # TODO: use a background thread for flushing
        self._instance_id = instance_id
        self._manager = manager
        self._events = list()   # type: List[ProfileEvent]

    def start(self, event_type: ProfileEventType, port: Optional[Port]=None,
              port_length: Optional[int]=None, slot: Optional[int]=None,
              message_size: Optional[int]=None
              ) -> ProfileEvent:
        """Start measuring an event.

        Call this, then call stop() on the returned ProfileEvent at
        the end of the event.

        Args:
            instance_id: Instance for which this event occurred.
            event_type: Type of event that occurred.
            port: Port that was sent or received on.
            port_length: Length of the port, if vector.
            slot: Slot that was sent or received on.
            message_size: Size in bytes of the message.

        Returns:
            A new ProfileEvent.
        """
        if len(self._events) >= 100:
            self.__flush()

        now = Timestamp(time())
        event = ProfileEvent(self._instance_id, now, now, event_type, port,
                             port_length, slot, message_size)
        self._events.append(event)
        return event

    def shutdown(self) -> None:
        self.__flush()

    def record_event(self, event: ProfileEvent) -> None:
        """Record a profiling event.

        This will record the event, and may flush this and previously
        recorded events to the manager.

        Args:
            event: The event to record.
        """
        self._events.append(event)
        if len(self._events) >= 100:
            self.__flush()
        if event.event_type == ProfileEventType.DEREGISTER:
            self.__flush()

    def __flush(self) -> None:
        if self._events:
            self._manager.submit_profile_events(self._events)
            self._events.clear()



from ymmsl import Reference

from libmuscle.mcp.message import Message


def test_create() -> None:
    sender = Reference('sender.port')
    receiver = Reference('receiver.port')
    timestamp = 10.0
    next_timestamp = 11.0
    settings_overlay = (6789).to_bytes(2, 'little', signed=True)
    data = (12345).to_bytes(2, 'little', signed=True)
    msg = Message(sender, receiver, None, timestamp, next_timestamp,
                  settings_overlay, data)
    assert msg.sender == sender
    assert msg.receiver == receiver
    assert msg.port_length is None
    assert msg.timestamp == 10.0
    assert msg.next_timestamp == 11.0
    assert msg.settings_overlay == settings_overlay
    assert msg.data == data



import datetime

import google.protobuf.timestamp_pb2 as pbts


class Timestamp:
    """A timestamp, as the number of seconds since the UNIX epoch.

    Args:
        seconds: The number of seconds since the start of 1970.
    """
    def __init__(self, seconds: float) -> None:
        self.__seconds = seconds

    def to_rfc3339(self) -> str:
        """Converts a Timestamp to a datetime string.

        Returns:
            The timestamp as a string.
        """
        date_time = datetime.datetime.utcfromtimestamp(self.__seconds)
        return date_time.isoformat() + 'Z'

    @staticmethod
    def from_grpc(timestamp: pbts.Timestamp) -> 'Timestamp':
        """Creates a Timestamp from a gRPC Timestamp message.

        Args:
            timestamp: A gRPC Timestamp from a gRPC call.

        Returns:
            The same timestamp as a Timestamp object.
        """
        return Timestamp(timestamp.seconds + timestamp.nanos * 1e-9)

    def to_grpc(self) -> pbts.Timestamp:
        """Converts a Timestamp to the gRPC type.

        Returns:
            The same timestamp, as a gRPC object.
        """
        seconds = int(self.__seconds)
        nanos = int((self.__seconds - seconds) * 10**9)
        return pbts.Timestamp(seconds=seconds, nanos=nanos)



import multiprocessing as mp
from pathlib import Path
import subprocess
from unittest.mock import MagicMock

import msgpack

from libmuscle.mcp.tcp_server import TcpServer
from libmuscle.mcp.message import Message
from libmuscle.post_office import PostOffice

from ymmsl import Reference, Settings


def tcp_server_process(control_pipe):
    control_pipe[0].close()
    settings = Settings({'test_setting': 42})
    data = {'test1': 10, 'test2': [None, True, 'testing']}
    receiver = Reference('test_receiver.test_port2')
    message = Message(
            Reference('test_sender.test_port'),
            receiver,
            10, 1.0, 2.0, settings, data).encoded()

    def get_message(receiver):
        assert receiver == 'test_receiver.test_port2'
        return message

    post_office = MagicMock()
    post_office.done = False
    post_office.get_message = get_message

    sender_instance_id = Reference('test_sender')
    server = TcpServer(sender_instance_id, post_office)
    control_pipe[1].send(server.get_location())
    control_pipe[1].recv()
    control_pipe[1].close()
    server.close()


def test_cpp_tcp_client(log_file_in_tmpdir):
    # create server process
    server_pipe = mp.Pipe()
    server_process = mp.Process(target=tcp_server_process, args=(server_pipe,))
    server_process.start()
    server_pipe[1].close()
    server_loc = server_pipe[0].recv()

    # create C++ client
    # it receives and checks settings, and sends a log message
    # see libmuscle/cpp/src/libmuscle/tests/mmp_client_test.cpp
    cpp_build_dir = Path(__file__).parents[1] / 'libmuscle' / 'cpp' / 'build'
    lib_paths = [
            cpp_build_dir / 'grpc' / 'c-ares' / 'c-ares' / 'lib',
            cpp_build_dir / 'grpc' / 'zlib' / 'zlib' / 'lib',
            cpp_build_dir / 'grpc' / 'openssl' / 'openssl' / 'lib',
            cpp_build_dir / 'protobuf' / 'protobuf' / 'lib',
            cpp_build_dir / 'grpc' / 'grpc' / 'lib',
            cpp_build_dir / 'msgpack' / 'msgpack' / 'lib']
    env = {
            'LD_LIBRARY_PATH': ':'.join(map(str, lib_paths))}
    cpp_test_dir = cpp_build_dir / 'libmuscle' / 'tests'
    cpp_test_client = cpp_test_dir / 'tcp_client_test'
    result = subprocess.run([str(cpp_test_client), server_loc], env=env)

    server_pipe[0].send(None)
    server_pipe[0].close()
    server_process.join()

    assert result.returncode == 0
    assert server_process.exitcode == 0
